local helper = wesnoth.require "lua/helper.lua"
local T = helper.set_wml_tag_metatable {}
-- local debug_utils = wesnoth.require "~add-ons/World_of_Wesnoth/lua/utils.lua"


function wesnoth.update_stats(unit)
	-- This is the main unit properties generator, from the key numbers, it generates all properties

	-- First, we get some properties we don't want to change during the modification
	local health_percentage = unit.hitpoints / unit.max_hitpoints
	local init_mana,init_max_mana = wesnoth.get_mana(unit)
	local mana_percentage = init_mana / init_max_mana
	local side_number = wesnoth.get_variable("side_number")
	local melee_damage_bonus = 0
	local ranged_damage_bonus = 0
	local original_id = unit.id
	local race, intelligence, strength, dexterity, endurance, total_level, melee_damage, melee_damage_plus, ranged_damage, ranged_damage_plus, melee_precision, ranged_precision, melee_attacks, melee_attacks_plus, ranged_attacks, ranged_attacks_plus, movement, resist_fire, resist_cold, resist_arcane, resist_blade, resist_impact, resist_pierce, skills, specials_melee, specials_ranged, defense, absorb, skills_modified, melee_suck, melee_suck_mana, ranged_suck, ranged_suck_mana, penetrate_fire, penetrate_cold, penetrate_arcane, penetrate_blade, penetrate_impact, penetrate_pierce, magic_damage, melee_type, ranged_type, mana_regen, hitpoints, courage, effects, alignment, class_levels, mana, weapon_type, str_req, int_req, dex_req = wesnoth.get_unit_properties(unit)	-- It was useful to separate this function, because these properties were needed also elsewhere
	raw_unit = wesnoth.create_unit { type="Hero", id=unit.id , gender=unit.gender, x=unit.x, y=unit.y, canrecruit=true, side=unit.side, moves=unit.moves, attacks_left=unit.attacks_left, name=unit.name, experience=unit.experience, { "variables" , properties }, {"modifications" , {{ "advance", id="basic_properties" ,effects}}}, { "abilities", {}}}	-- To reset everything, the unit is created from scratch
	unit = raw_unit.__cfg
	unit.id = original_id
	-- We find the class with the lowest base bonus to hitpoints, so the multiclassers will be somewhat weakened by this
	local lowest_bonus = 1000	-- All classes have less than 1000 starting hitpoints, so the first class' starting hitpoints should replace it
	for i,j in pairs(class_levels) do
		if classes[i][2].hitpoints < lowest_bonus then
			lowest_bonus = classes[i][2].hitpoints
		end
		unit.max_hitpoints = unit.max_hitpoints + classes[i][2].hitpoints_per_level * j
		unit.max_attacks = unit.max_attacks + ( (j / classes[i][2].attack_per) - (j / classes[i][2].attack_per) % 1)
	end
	if lowest_bonus == 1000 then
		helper.wml_error("Failed to find a class to determine hitpoints.")
	end
	-- Setup hitpoints and level
	unit.max_hitpoints = unit.max_hitpoints + endurance * total_level * 0.25 + hitpoints + lowest_bonus
	unit.hitpoints = unit.max_hitpoints * health_percentage
	unit.level = total_level
	-- Setup mana, we have to put it into variables, because it would be erased if toplevel
	for i = 1,#unit do
		if unit[i][1] == "variables" then
			unit[i][2].max_mana = intelligence + mana
			unit[i][2].mana = mana_percentage * unit[i][2].max_mana
			unit[i][2].mana_regen = 1 + (intelligence - intelligence % 10) / 8 + mana_regen
			-- Setup maximum numbers of summons
			for j = 1,#unit[i][2] do
				if unit[i][2][j][1] == "summons" then
					table.remove(unit[i][2],j)
					break
				end
			end
			local summons = {}
			if skills_modified[112] then
				summons.summon_112_max = 1
			end
			if skills_modified[214] then
				summons.summon_214_max = 1
			end
			if skills_modified[305] then
				summons.summon_305_max = 1
			end
			if skills_modified[403] then
				summons.summon_403_max = 1
			end
			if skills_modified[410] then
				summons.summon_410_max = 100	-- Let's assume that this means infinity
			end
			if skills_modified[512] then
				summons.summon_512_max = 1
			end
			if skills_modified[607] then
				summons.summon_607_max = 1
			end
			if skills_modified[702] then
				summons.summon_702_max = skills_modified[702]
			end
			if skills_modified[706] then
				summons.summon_706_max = skills_modified[706]
			end
			if skills_modified[713] then
				summons.summon_713_max = skills_modified[713]
			end
			if skills_modified[716] then
				summons.summon_716_max = skills_modified[716]
			end
			table.insert(unit[i][2],{ "summons" , summons})
		end
	end
	if skills_modified[207] and weapon_type == "sword" then
		melee_damage_bonus = melee_damage_bonus + (1 + skills_modified[207])	-- Fencer's sword mastery
	end
	if skills_modified[210] and weapon_type == "rapier" then
		melee_damage_bonus = melee_damage_bonus + (1 + skills_modified[207])	-- Fencer's rapier mastery
	end
	if skills_modified[213] and (weapon_type == "rapier" or weapon_type == "dagger" or weapon_type == "sword") then
		melee_damage_bonus = melee_damage_bonus + (1 + skills_modified[207]) * 0.05 * dexterity	-- Fencer's dextrous combat
	end
	if skills_modified[314] then
		melee_damage_bonus = melee_damage_bonus - skills_modified[314]		-- Hunter's caution
	end
	for i=1,#unit do
		if unit[i][1]=="attack" and unit[i][2].name == "default_attack" then
			if skills_modified[404] then	-- We have to check if 404 will be found, Rogue's planning skill applies here
				unit[i][2].damage = (unit[i][2].damage + (strength/5) + (intelligence * (1 + skills_modified[404]) * 0.07) + melee_damage_bonus) * ( 100 + melee_damage) / 100 + melee_damage_plus
			else
				unit[i][2].damage = (unit[i][2].damage + (strength/5) + melee_damage_bonus) * ( 100 + melee_damage) / 100 + melee_damage_plus
			end
			if skills_modified[709] then
				unit[i][2].damage = unit[i][2].damage + ( 1 + skills_modified[709]) * endurance / 15	-- Tough joints applying here
			end
			if skills_modified[2] then
				unit[i][2].damage = unit[i][2].damage + skills_modified[2] * 2		-- Aggression takes effect here
			end
			unit[i][2].number = unit[i][2].number * ( 100 + melee_attacks + dexterity ^ 0.6 * 15) / 100 + melee_attacks_plus
			unit[i][2].type = melee_type
			for j=1,#unit[i][2] do
				if unit[i][2][j][1] == "specials" then
					-- Some weapon modifiers apply as weapon specials. some of them like suck are event-based, but the events read their properties
					for k=1,#unit[i][2][j][2] do
						if unit[i][2][j][2][k][1] == "chance_to_hit" then
							if unit[i][2][j][2][k][2].id == "precision" then
								unit[i][2][j][2][k][2].add = melee_precision + dexterity ^ 0.7
							end
						end
					end
					if melee_suck > 0 then
						table.insert(unit[i][2][j][2],{ "suck" , { id="suck" , value=melee_suck  }})
					end
					if melee_suck_mana > 0 then
						table.insert(unit[i][2][j][2],{ "suck_mana" , { id="suck_mana" , value=melee_suck_mana  }})
					end
					if absorb > 0 then
						table.insert(unit[i][2][j][2],{ "damage" , { id="absorb" , sub=absorb, apply_to="opponent" }})
					end
					if #specials_melee > 0 then
						if #specials_melee[1] > 0 then
							for k = 1,#specials_melee[1] do
								table.insert(unit[i][2][j][2],specials_melee[1][k])
							end
						end
					end
				end
			end
			wesnoth.set_variable("melee_attack",unit[i][2]) -- This will let us make a deepcopy anytime it is needed, and relatively quickly
		end
		if unit[i][1]=="attack" and unit[i][2].name == "default_ranged_attack" then
			unit[i][2].damage = (unit[i][2].damage + (strength/4) + ranged_damage_bonus) * ( 100 + ranged_damage) / 100 + ranged_damage_plus
			unit[i][2].number = unit[i][2].number * ( 100 + ranged_attacks + dexterity ^ 0.7 * 15) / 100 + ranged_attacks_plus
			unit[i][2].type = ranged_type
			for j=1,#unit[i][2] do
				if unit[i][2][j][1] == "specials" then
					-- Some weapon modifiers apply as weapon specials. some of them like suck are event-based, but the events read their properties
					for k=1,#unit[i][2][j][2] do
						if unit[i][2][j][2][k][1] == "chance_to_hit" then
							if unit[i][2][j][2][k][2].id == "precision" then
								unit[i][2][j][2][k][2].add = ranged_precision + dexterity ^ 0.8
							end
						end
					end
					if ranged_suck > 0 then
						table.insert(unit[i][2][j][2],{ "suck" , { id="suck" , value=melee_suck  }})
					end
					if ranged_suck_mana > 0 then
						table.insert(unit[i][2][j][2],{ "suck_mana" , { id="suck_mana" , value=melee_suck_mana  }})
					end
					if absorb > 0 then
						table.insert(unit[i][2][j][2],{ "damage" , { id="absorb" , sub=absorb, apply_to="opponent" }})
					end
					if #specials_ranged > 0 then
						if #specials_ranged[1] > 0 then
							for k = 1,#specials_ranged[1] do
								table.insert(unit[i][2][j][2],specials_ranged[1][k])
							end
						end
					end
				end
			end
			wesnoth.set_variable("ranged_attack",unit[i][2])
		end
	end
	unit.max_experience = wesnoth.calculate_max_experience(total_level)	-- For easier access and modification, the function that calculates the experience required to level-up is on a single place
	-- Make some properties take place here
	unit.alignment = alignment
	unit.race = race
	-- Some properties add traits, 
	if courage then
		for j = 1,#unit do
			if unit[j][1] == "modifications" then
				table.insert(unit[j][2],{ "trait" , { id="fearless" }})
			end
		end
	end
	if #effects > 0 then
		for j = 1,#unit do
			if unit[j][1] == "modifications" then
				for k = 1,#effects do
					-- Effects are added in [advance] tags, in order to avoid interfering with WML, that might use [object] tags, but not AMLA
					table.insert(unit[j][2], { "advance" , { effects[k]}})
				end
			end
		end
	end
	-- Resistance penetrations apply as aurae that lower the resistances of all nearby enemies
	if penetrate_fire > 0 then
		for j = 1,#unit do
			if unit[j][1] == "abilities" then
				table.insert(unit[j][2],{ "resistance" , { id="penetrate_fire" , sub=penetrate_fire , max_value = 80 , apply_to = "fire", affect_enemies=true, affect_allies=false, affect_self=false, { "affect_adjacent" , { adjacent="n,ne,se,s,sw,nw"  }} }})
			end
		end	
	end
	if penetrate_cold > 0 then
		for j = 1,#unit do
			if unit[j][1] == "abilities" then
				table.insert(unit[j][2],{ "resistance" , { id="penetrate_cold" , sub=penetrate_cold , max_value = 80 , apply_to = "cold", affect_enemies=true, affect_allies=false, affect_self=false, { "affect_adjacent" , { adjacent="n,ne,se,s,sw,nw"  }} }})
			end
		end	
	end
	if penetrate_arcane > 0 then
		for j = 1,#unit do
			if unit[j][1] == "abilities" then
				table.insert(unit[j][2],{ "resistance" , { id="penetrate_arcane" , sub=penetrate_arcane , max_value = 80 , apply_to = "arcane", affect_enemies=true, affect_allies=false, affect_self=false, { "affect_adjacent" , { adjacent="n,ne,se,s,sw,nw"  }} }})
			end
		end	
	end
	if penetrate_blade > 0 then
		for j = 1,#unit do
			if unit[j][1] == "abilities" then
				table.insert(unit[j][2],{ "resistance" , { id="penetrate_blade" , sub=penetrate_blade , max_value = 80 , apply_to = "blade", affect_enemies=true, affect_allies=false, affect_self=false, { "affect_adjacent" , { adjacent="n,ne,se,s,sw,nw"  }} }})
			end
		end	
	end
	if penetrate_pierce > 0 then
		for j = 1,#unit do
			if unit[j][1] == "abilities" then
				table.insert(unit[j][2],{ "resistance" , { id="penetrate_pierce" , sub=penetrate_pierce , max_value = 80 , apply_to = "pierce", affect_enemies=true, affect_allies=false, affect_self=false, { "affect_adjacent" , { adjacent="n,ne,se,s,sw,nw"  }} }})
			end
		end	
	end
	if penetrate_impact > 0 then
		for j = 1,#unit do
			if unit[j][1] == "abilities" then
				table.insert(unit[j][2],{ "resistance" , { id="penetrate_impact" , sub=penetrate_impact , max_value = 80 , apply_to = "impact", affect_enemies=true, affect_allies=false, affect_self=false, { "affect_adjacent" , { adjacent="n,ne,se,s,sw,nw"  }} }})
			end
		end	
	end
	local cth_modifier = 0
	local fire_mastery = 0
	if skills_modified[617] then
		fire_mastery=skills_modified[617]*15
	end
	local cold_mastery = 0
	if skills_modified[616] then
		cold_mastery=skills_modified[616]*15
	end
	for i=1,#unit do	-- This is to allow WML and some lua to easily read the skill values, just by using unit.variables.skill 101 for example to get the level of thrash (or empty if the character does not have it)
		if unit[i][1] == "variables" then
			for j,k in pairs(skills_modified) do
				unit[i][2]["skill_" .. tostring(j)] = k
			end
			if unit[i][2].skill_602 then
				if not unit[i][2].skill_208 then
					unit[i][2].skill_208 = 0
				end
				unit[i][2].skill_208 = unit[i][2].skill_602 + unit[i][2].skill_208
			end
			unit[i][2].magic_damage = magic_damage
			unit[i][2].fire_mastery = fire_mastery
			unit[i][2].cold_mastery = cold_mastery
		end
	end
	if skills_modified[1] then		-- Faerie fire for elves
		local damage = (4 + (skills_modified[1] ^ 2)) * (1 + magic_damage / 100)
		local current_attack = { name="faerie fire", description="faerie fire", range="ranged", icon="attacks/faerie-fire.png", damage=damage, number=3, type="arcane", {"specials" , { { "chance_to_hit" , { id="magical", name="magical", description="This attack always has a 70% chance to hit.", value=70, cumulative="no"  } } }} }
		unit[#unit + 1] = { "attack" , current_attack}
	end
	-- Skills that add attacks are added here, it is somewhat repetitive, so I will describe it only here. It usually contains a variable for computing some properties and then it adds a WML block.
	-- Skill 102 means second skill of first class, and that is Bash.
	if skills_modified[101] then
		local current_attack = wesnoth.get_variable("melee_attack")
		current_attack.damage = current_attack.damage * (1.2 + 0.2 * skills_modified[101])
		current_attack.defense_weight = 0
		current_attack.description = "thrash"
		local percentage = 30 + 2 * skills_modified[101]
		for i = 1,#current_attack do
			if current_attack[i][1] == "specials" then
				table.insert(current_attack[i][2], { "damage", { id="thrash" , name="thrash", description="Thrash: The attacker using this attack takes " .. tostring(percentage) .. "% more damage." , multiply = 1 + percentage/100, apply_to="opponent"  } } )
			end
		end
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[102] then
		local current_attack = wesnoth.get_variable("melee_attack")
		current_attack.damage = current_attack.damage * (0.9 + 0.1 * skills_modified[102])
		current_attack.defense_weight = 0
		current_attack.description = "bash"
		for i = 1,#current_attack do
			if current_attack[i][1] == "specials" then
				table.insert(current_attack[i][2], { "dummy", { id="knockback" , name="knockback", description="Knockback: This attack knocks enemies back." } } )
			end
		end
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[103] then
		-- Some skills can affect also resistances
		resist_blade = resist_blade + 4 + 2 * skills_modified[103]
		resist_impact = resist_impact + 4 + 3 * skills_modified[103]
		resist_pierce = resist_pierce + 4 + 2 * skills_modified[103]
	end
	if skills_modified[104] then
		local damage = (7 + 2 * (skills_modified[104] + skills_modified[104] ^ 2)) * (1 + magic_damage / 100)
		local current_attack = { name="spiritual strike", description="spiritual strike", range="spell", icon="attacks/magic-missile.png", damage=damage, number=1, type="arcane", defense_weight=0, {"specials" , { { "chance_to_hit" , { id="spell", name="spell", mana=4, description="Spell: This attack is a spell, so it costs 4 mana and has a 100% chance to hit.", value=100, cumulative="no"  } } }} }
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[105] then
		resist_fire = resist_fire + (skills_modified[105] + 1) * (endurance / 8)
		resist_cold = resist_cold + (skills_modified[105] + 1) * (endurance / 8)
		resist_arcane = resist_arcane + (skills_modified[105] + 1) * (endurance / 8)
	end
	if skills_modified[106] then
		local current_attack = wesnoth.get_variable("melee_attack")
		current_attack.damage = current_attack.damage * (0.85 + 0.15 * skills_modified[106])
		current_attack.defense_weight = 0
		current_attack.description = "fury"
		local percentage = 8 + 2 * skills_modified[106]
		for i = 1,#current_attack do
			if current_attack[i][1] == "specials" then
				table.insert(current_attack[i][2], { "chance_to_hit", { id="fury" , name="fury", description="Fury: Whenever this attack hits, it has a 50% chance to hit again, but the attacker's chance to get hit is increased by " .. tostring(percentage) .. "%." , add=percentage, apply_to="opponent", cumulative="yes"  } } )
			end
		end
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[108] then
		local current_attack = wesnoth.get_variable("melee_attack")
		current_attack.damage = current_attack.damage * current_attack.number * (0.85 + 0.15 * skills_modified[108])
		current_attack.number = 1
		current_attack.defense_weight = 0
		current_attack.description = "concentrate"
		local percentage = 78 + 2 * skills_modified[108]
		for i = 1,#current_attack do
			if current_attack[i][1] == "specials" then
				table.insert(current_attack[i][2], { "chance_to_hit", { id="concentrate" , name="concentrate", description="Concentrate: All the damage of this attack is done in a single blow, with a " .. tostring(percentage) .. "% chance to hit." , value=percentage, apply_to="self", cumulative="yes"  } } )
			end
		end
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[110] then
		local current_attack = wesnoth.get_variable("melee_attack")
		current_attack.attacks = math.ceil(current_attack.attacks * 0.5)
		current_attack.defense_weight = 0
		current_attack.description = "mayhem"
		for i = 1,#current_attack do
			if current_attack[i][1] == "specials" then
				table.insert(current_attack[i][2], { "dummy", { id="mayhem" , name="mayhem", description="Mayhem: The target's physical damage will be decreased by " .. skills_modified[110] .. "%. Stacks with itself. Lasts for 6 turns in average."  } } )
			end
		end
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[111] then
		for j = 1,#unit do
			if unit[j][1] == "abilities" then
				table.insert(unit[j][2],{ "regenerate" , { id="regrowth" , name="regrowth" , description="regrowth: Heals " .. tostring(16 + skills_modified[111]*8) .. " hitpoints per turn, or heals poison instead." , value=16 + skills_modified[111]*8 , affect_self = true, poison = "cured"}})
			end
		end	
	end
	if skills_modified[113] then
		local current_attack = wesnoth.get_variable("melee_attack")
		current_attack.damage = current_attack.damage * (1 + 0.2 * skills_modified[113])
		current_attack.description = "doom"
		for i = 1,#current_attack do
			if current_attack[i][1] == "specials" then
				table.insert(current_attack[i][2], { "damage", { id="doom" , name="doom", description="Doom: The opponent retaliates 50% less damage." , divide=2, apply_to="opponent", cumulative="yes"  } } )
			end
		end
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[114] then
		for j = 1,#unit do
			if unit[j][1] == "abilities" then
				table.insert(unit[j][2],{ "dummy" , { id="bloodlust" , name="bloodlust" , description="bloodlust: Heals " .. tostring(8 + skills_modified[114]*8) .. " hitpoints after each kill." }})
			end
		end	
	end
	if skills_modified[115] then
		local current_attack = wesnoth.get_variable("melee_attack")
		current_attack.damage = current_attack.damage * (0.5 + 0.1 * skills_modified[115])
		current_attack.defense_weight = 0
		current_attack.description = "shockwave"
		for i = 1,#current_attack do
			if current_attack[i][1] == "specials" then
				table.insert(current_attack[i][2], { "dummy", { id="shockwave" , name="shockwave", description="Shockwave: " .. tostring(10 + 15 * skills_modified[115]) .. "% of the damage will be taken by all adjacent targets as impact damage."  } } )
			end
		end
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[117] then
		local current_attack = wesnoth.get_variable("melee_attack")
		current_attack.damage = current_attack.damage * (1.2 + 0.2 * skills_modified[117])
		current_attack.defense_weight = 0
		current_attack.description = "cleave"
		for i = 1,#current_attack do
			if current_attack[i][1] == "specials" then
				table.insert(current_attack[i][2], { "dummy", { id="cleave" , name="cleave", description="Cleave: deals a half of the damage also to targets next to both attacker and defender."  } } )
			end
		end
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[201] then
		local current_attack = wesnoth.get_variable("melee_attack")
		current_attack.damage = current_attack.damage * (0.85 + 0.15 * skills_modified[201])
		current_attack.defense_weight = 0
		current_attack.description = "confusing attack"
		for i = 1,#current_attack do
			if current_attack[i][1] == "specials" then
				table.insert(current_attack[i][2], { "attacks", { id="confusing attack" , name="confusing attack", description="Confusing Attack: The target will retaliate one attack less." , sub=1, apply_to="opponent", cumulative="yes"  } } )
			end
		end
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[202] then
		resist_fire = resist_fire + (skills_modified[202] + 1) * (dexterity / 7)
		resist_cold = resist_cold + (skills_modified[202] + 1) * (dexterity / 7)
		resist_arcane = resist_arcane + (skills_modified[202] + 1) * (dexterity / 7)
	end
	if skills_modified[204] then
		local current_attack = wesnoth.get_variable("melee_attack")
		current_attack.damage = current_attack.damage * (0.9 + 0.1 * skills_modified[204])
		current_attack.defense_weight = 0
		current_attack.description = "bleeding strike"
		for i = 1,#current_attack do
			if current_attack[i][1] == "specials" then
				table.insert(current_attack[i][2], { "dummy", { id="bleeding" , name="bleeding", description="Bleeding: The target will start taking pierce damage of" .. 8 + skills_modified[204] * 4 .. "points per turn. Lasts for 6 turns in average."  } } )
			end
		end
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[205] then
		-- It marks that the chance to get hit is modified here
		cth_modifier = cth_modifier + (1 + skills_modified[205]) * strength / 10
	end
	if skills_modified[2] then
		-- Aggression increases the chance to get hit
		cth_modifier = cth_modifier - skills_modified[2] * 2
	end
	if skills_modified[314] then
		cth_modifier = cth_modifier + skills_modified[314] * 2
	end
	if skills_modified[206] then
		local current_attack = wesnoth.get_variable("melee_attack")
		current_attack.damage = current_attack.damage * (0.8 + 0.2 * skills_modified[206])
		current_attack.defense_weight = 0
		current_attack.description = "desintegration"
		for i = 1,#current_attack do
			if current_attack[i][1] == "specials" then
				table.insert(current_attack[i][2], { "dummy", { id="desintegration" , name="desintegration", description="Desintegration: The target's physical resistances will be decreased by " .. 5 + skills_modified[206] .. "%. Stacks with itself. Lasts for 6 turns in average."  } } )
			end
		end
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[602] then
		-- Both Wizard and Fencer have fireball, so it is summed here.
		if skills_modified[208] then
			skills_modified[208] = skills_modified[208] + skills_modified[602]
		else
			skills_modified[208] = skills_modified[602]
		end
	end
	if skills_modified[208] then
		local damage = (7 + 2 * (2 * skills_modified[208] + skills_modified[208] ^ 2)) * (1 + (magic_damage + fire_mastery) / 100)
		local current_attack = { name="fireball", description="fireball", range="spell", icon="attacks/fireball.png", damage=damage, number=1, type="fire", defense_weight=0, {"specials" , { { "chance_to_hit" , { id="spell", name="spell", mana=5, description="Spell: This attack is a spell, so it costs 5 mana and has a 100% chance to hit.", value=100, cumulative="no"  } } }} }
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[209] then
		local current_attack = wesnoth.get_variable("melee_attack")
		current_attack.damage = current_attack.damage * (0.9 + 0.1 * skills_modified[209])
		current_attack.description = "fence"
		local percentage = 34 - skills_modified[209] * 4
		for i = 1,#current_attack do
			if current_attack[i][1] == "specials" then
				table.insert(current_attack[i][2], { "chance_to_hit", { id="fence" , name="fence", description="Fence: This attack's chance to hit is decreased by " .. tostring(percentage) .. "%, but the opponent's counterattack's chance to hit is divided by three." , sub=percentage, apply_to="self", cumulative="yes"  } } )
				table.insert(current_attack[i][2], { "chance_to_hit", { id="fence-second" , divide=3 , apply_to="opponent", cumulative="yes"  } } )
			end
		end
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[212] then
		local current_attack = wesnoth.get_variable("melee_attack")
		current_attack.number = current_attack.number * (0.4 + 0.1 * skills_modified[212])
		current_attack.defense_weight = 0
		current_attack.description = "parry"
		for i = 1,#current_attack do
			if current_attack[i][1] == "specials" then
				table.insert(current_attack[i][2], { "dummy", { id="parry" , name="parry", description="Parry: If the opponent misses a retaliation attack, the attacker has a " .. tostring(40 + 10 * skills_modified[212]) .. "% chance to hit him for free."  } } )
			end
		end
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[215] then
		local current_attack = wesnoth.get_variable("melee_attack")
		current_attack.number = math.ceil(current_attack.number * ( 0.4 + skills_modified[215]))
		current_attack.defense_weight = 0
		current_attack.range="ranged"
		current_attack.description = "swordmerang"
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[217] then
		local current_attack = wesnoth.get_variable("melee_attack")
		current_attack.number = math.ceil(current_attack.number / 2)
		current_attack.damage = current_attack.damage * (0.8 + 0.2 * skills_modified[217])
		current_attack.defense_weight = 0
		current_attack.description = "whirlwind"
		local value = 2 * skills_modified[217]
		for i = 1,#current_attack do
			if current_attack[i][1] == "specials" then
				table.insert(current_attack[i][2], { "attacks", { id="whirlwind" , name="whirlwind", description="Whirlwind: Deals damage to all adjacent targets, enemies cannot retaliate, chance to hit increased by " .. tonstring(value) .. "%." , value=0, apply_to="opponent" } } )
				table.insert(current_attack[i][2], { "chance_to_hit", { id="whirlwind" , add = value, apply_to="self" } } )
			end
		end
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[301] then
		local current_attack = wesnoth.get_variable("ranged_attack")
		current_attack.damage = current_attack.damage * current_attack.number * (0.45 + 0.15 * skills_modified[301])
		current_attack.name = "long_range_ranged_attack"
		current_attack.number = 1
		current_attack.defense_weight = 0
		current_attack.description = "long range shot"
		for i = 1,#current_attack do
			if current_attack[i][1] == "specials" then
				table.insert(current_attack[i][2], { "attacks", { id="long_range_shot" , name="long range shot", description="Long range shot: This attack cannot be countered", value=0, apply_to="opponent"} } )
				table.insert(current_attack[i][2], { "chance_to_hit", { id="magical" , name="magical", description="Magical: This attacks has always a 70% chance to hit, regardless of the target unit's defensive ability." , value=70, apply_to="self", cumulative="no"  } } )
			end
		end
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[302] then
		local current_attack = wesnoth.get_variable("melee_attack")
		current_attack.damage = current_attack.damage * (1.2 + 0.2 * skills_modified[302])
		current_attack.description = "struggle"
		local percentage = 1.2 + 0.2 * skills_modified[302]
		for i = 1,#current_attack do
			if current_attack[i][1] == "specials" then
				table.insert(current_attack[i][2], { "chance_to_hit", { id="struggle" , name="struggle", description="Struggle: This attack's chance to hit is decreased by 10%, but its damage is increased by " .. tostring(percentage*100) .. "%." , sub=10, apply_to="self", cumulative="yes"  } } )
			end
		end
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[304] then
		unit.max_moves = unit.max_moves + 1
		if skills_modified[304] >= 2 then
			for i=1,#unit do
				if unit[i][1] == "movement_costs" then
					local function decrease_cost(terrain_name)
						if unit[i][2][terrain_name] then
							if unit[i][2][terrain_name] > 1 then
								unit[i][2][terrain_name] = unit[i][2][terrain_name] - 1
							end
						end
					end
					decrease_cost("forest")
					decrease_cost("hills")
					decrease_cost("mountains")
					decrease_cost("fungus")
					decrease_cost("frozen")
					if skills_modified[304] >= 3 then
						unit.max_moves = unit.max_moves + 1
					end
					if skills_modified[304] >= 4 then
						decrease_cost("forest")
						decrease_cost("hills")
						decrease_cost("mountains")
						decrease_cost("fungus")
						decrease_cost("frozen")
						decrease_cost("shallow_water")
						decrease_cost("deep_water")
						decrease_cost("coastal_reef")
						decrease_cost("swamp")
					end
					if skills_modified[304] >= 5 then
						unit.max_moves = unit.max_moves + skills_modified[304] - 4
					end
				end
			end
		end
	end
	if skills_modified[306] then
		local current_attack = wesnoth.get_variable("ranged_attack")
		current_attack.defense_weight = 0
		current_attack.description = "penetrating shot"
		local percentage = 0.4 + 0.1 * skills_modified[306]
		if percentage > 1 then
			percentage = 1
			current_attack.damage = current_attack.damage * ((skills_modified[306] - 6) * 0.2 + 1)
		end
		for i = 1,#current_attack do
			if current_attack[i][1] == "specials" then
				table.insert(current_attack[i][2], { "dummy", { id="penetrating" , name="penetrating", description="Penetrating: This attack will do " .. tostring(percentage*100) .. "% of its damage to the target behind it, and repeat this effect." , value = percentage } } )
			end
		end
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[308] then
		local current_attack = wesnoth.get_variable("ranged_attack")
		current_attack.damage = current_attack.damage * current_attack.number * (0.80 + 0.2 * skills_modified[308])
		current_attack.name = "quick_ranged_attack"
		current_attack.number = 1
		current_attack.defense_weight = 0
		current_attack.description = "rapid shot"
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[309] then
		-- Some skills add also abilities
		for i=1,#unit do
			if unit[i][1] == "abilities" then
				local pattern = "*^Fds"
				if skills_modified[309] >= 2 then
					pattern = pattern .. ",*^Fet,*^Fms,*^Fmf,*^Fdf"
				end
				if skills_modified[309] >= 3 then
					pattern = pattern .. ",*^Fetd,*^Ft,*^Fdw,*^Fmv"
				end
				if skills_modified[309] >= 4 then
					pattern = pattern .. ",*^Fda,*^Fma,*^Fpa,*^Uf"
				end
				if skills_modified[309] >= 5 then
					pattern = pattern .. ",*^Ufi,Uh*,Mm*,Ms*,Md*"
				end
				if skills_modified[309] >= 6 then
					pattern = pattern .. ",Hh*,Hhd*,Hd*,Ha*,*^Dr"
				end
				if skills_modified[309] >= 7 then
					pattern = pattern .. "S*,WWr*"
				end
				table.insert(unit[i][2], { "hides", { id="stalk" , name="stalk", name_inactive="stalk", description="Stalk: This unit can hide on a variety of terrains, depending on the skill level, on low levels, it can hide only in deciduous forests, higher skills allow higing also in coniferous forests, autumn and winter forests, on really high levels also in hills and possible even in swamps." , affect_self=true , { "filter_self", {  { "filter_location" , {   terrain=pattern } }  } } } } )

			end
		end
	end
	if skills_modified[310] then
		local current_attack = wesnoth.get_variable("melee_attack")
		current_attack.description = "cautious attack"
		local percentage = 0.8 - 0.1 * skills_modified[310]
		for i = 1,#current_attack do
			if current_attack[i][1] == "specials" then
				table.insert(current_attack[i][2], { "chance_to_hit", { id="struggle" , name="struggle", description="Cautious attack: This attack's chance to hit is decreased by 10%, but the opponent's chance to hit is only " .. tostring(percentage*100) .. "% of its original value." , sub=10, apply_to="self", cumulative="yes"  } } )
				table.insert(current_attack[i][2], { "chance_to_hit", { id="struggle" , multiply=percentage, apply_to="opponent", cumulative="yes"  } } )
			end
		end
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[313] then
		local current_attack = wesnoth.get_variable("ranged_attack")
		current_attack.damage = current_attack.damage * (1 + 0.05 * skills_modified[313])
		current_attack.defense_weight = 0
		current_attack.description = "stunning shot"
		local percentage = 5 * skills_modified[313]
		for i = 1,#current_attack do
			if current_attack[i][1] == "specials" then
				table.insert(current_attack[i][2], { "chance_to_hit", { id="stunning shot" , name="stunning shot", description="Stunning Shot: This attack's chance to hit is increased by " .. tostring(percentage) .. "%, and when hits, the opponent will not hit anymore until the end of fight." , add = percentage, apply_to="self", cumulative="yes"  } } )
			end
		end
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[315] then
		local current_attack = wesnoth.get_variable("ranged_attack")
		current_attack.damage = current_attack.damage * (1.6 + 0.3 * skills_modified[315])
		current_attack.defense_weight = 0
		current_attack.description = "anger"
		for i = 1,#current_attack do
			if current_attack[i][1] == "specials" then
				table.insert(current_attack[i][2], { "chance_to_hit", { id="anger" , name="anger", description="Anger: This attack's chance to hit is decreased by 10%, opponent chance to hit is increased by 10% and the attacker takes 25% more damage." , sub=10, apply_to="self", cumulative="yes"  } } )
				table.insert(current_attack[i][2], { "chance_to_hit", { id="anger" , add=10, apply_to="opponent", cumulative="yes"  } } )
				table.insert(current_attack[i][2], { "damage", { id="anger" , multiply=1.25, apply_to="opponent", cumulative="yes"  } } )
			end
		end
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[316] then
		local damage = (22 + 3 * (skills_modified[316] ^ 2)) * (1 + magic_damage / 100)
		local current_attack = { name="earthquake", description="earthquake", range="spell", icon="attacks/crush-wose.png", damage=damage, number=1, type="impact", defense_weight=0, {"specials" , { { "chance_to_hit" , { id="spell", name="spell", mana=11, description="Spell: This attack is a spell, so it costs 11 mana and has a 100% chance to hit.", value=100, cumulative="no"  } } }} }
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[317] then
		local current_attack = wesnoth.get_variable("ranged_attack")
		current_attack.defense_weight = 0
		current_attack.description = "devastating strafe"
		current_attack.damage = current_attack.damage * (0.8 + 0.2 * skills_modified[317])
		for i = 1,#current_attack do
			if current_attack[i][1] == "specials" then
				table.insert(current_attack[i][2], { "dummy", { id="devastating strafe" , name="devastating strafe", description="Devastating Strafe: This attack hits also enemies next to both attacker and defender and also unit behind them." } } )
			end
		end
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[401] then
		local current_attack = wesnoth.get_variable("melee_attack")
		current_attack.defense_weight = 0
		current_attack.description = "poisonous blow"
		if skills_modified[401] > 5 then
			current_attack.damage = current_attack.damage * (skills_modified[401] * 0.2)
		end
		if skills_modified[401] == 1 then
			for i = 1,#current_attack do
				if current_attack[i][1] == "specials" then
					table.insert(current_attack[i][2], { "dummy", { id="venom" , name="venom", description="Venom: This attack inflicts a status effect that deals 16 damage per turn and lasts for 6 turns in average. Does not stack with itself. Does not affect undead." } } )
				end
			end
		end
		if skills_modified[401] == 2 then
			for i = 1,#current_attack do
				if current_attack[i][1] == "specials" then
					table.insert(current_attack[i][2], { "dummy", { id="toxin" , name="toxin", description="Toxin: This attack inflicts a status effect that deals 24 damage per turn and lasts for 8 turns in average. Does not stack with itself. Does not affect undead."} } )
				end
			end
		end
		if skills_modified[401] == 3 then
			for i = 1,#current_attack do
				if current_attack[i][1] == "specials" then
					table.insert(current_attack[i][2], { "dummy", { id="acid" , name="acid", description="Acid: This attack inflicts a status effect that deals 32 damage per turn and lasts for 6 turns in average. Does not stack with itself. Does affect undead." } } )
				end
			end
		end
		if skills_modified[401] == 4 then
			for i = 1,#current_attack do
				if current_attack[i][1] == "specials" then
					table.insert(current_attack[i][2], { "dummy", { id="miasma" , name="miasma", description="Miasma: This attack inflicts a status effect that deals 40 damage per turn and lasts for 8 turns in average. Does not stack with itself. Does affect undead." } } )
				end
			end
		end
		if skills_modified[401] > 4 then
			for i = 1,#current_attack do
				if current_attack[i][1] == "specials" then
					table.insert(current_attack[i][2], { "dummy", { id="bane" , name="bane", description="Bane: This attack inflicts a status effect that deals 48 damage per turn and lasts for 11 turns in average. Does not stack with itself. Does affect undead." } } )
				end
			end
		end
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[402] then
		local current_attack = wesnoth.get_variable("melee_attack")
		current_attack.defense_weight = 0
		current_attack.description = "backstab"
		local percentage = 1 + 0.5 * skills_modified[402]
		for i = 1,#current_attack do
			if current_attack[i][1] == "specials" then
				local had_backstab = false
				for j = 1,#current_attack[i][2] do
					if current_attack[i][2][j][1] == "damage" then
						if current_attack[i][2][j][2].id == "backstab" then
							current_attack[i][2][j][2].multiply = current_attack[i][2][j][2].multiply + percentage - 1
							current_attack[i][2][j][2].description = "Backstab: When used offensively, this attack deals " .. tostring(current_attack[i][2][j][2].multiply * 100 - 100 + percentage * 100) .. "% damage if there is an enemy of the target on the opposite side of the target."
							had_backstab = true
						end
					end
				end
				if had_backstab == false then
					table.insert(current_attack[i][2], { "damage", { id="backstab" , name="backstab", description="Backstab: When used offensively, this attack deals " .. tostring(percentage * 100) .. "% damage if there is an enemy of the target on the opposite side of the target." , backstab = true, multiply = percentage } } )
				end
			end
		end
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[405] then
		local current_attack = wesnoth.get_variable("melee_attack")
		current_attack.damage = current_attack.damage * (0.7 + 0.5 * skills_modified[405])
		current_attack.number = 1
		current_attack.defense_weight = 0
		current_attack.description = "crippling strike"
		for i = 1,#current_attack do
			if current_attack[i][1] == "specials" then
				table.insert(current_attack[i][2], { "dummy", { id="crippling strike" , name="crippling strike", description="Crippling Strike: The target's damage will be decreased by 2. Stacks with itself. Lasts for 6 turns in average."  } } )
			end
		end
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[406] then
		local current_attack = { name="toxic cloud", description="toxic cloud", range="spell", icon="attacks/curse.png~GS()~CS(-50,0,-50)", damage=1, number=0, type="arcane", defense_weight=0, {"specials" , { { "chance_to_hit" , { id="spell", name="spell", mana=6, description="Spell: This attack is a spell, so it costs 6 mana and has a 100% chance to hit.", value=100, cumulative="no"  } }, { "dummy", { id="toxin" , name="toxin", description="Toxin: This attack inflicts a status effect that deals 24 damage per turn and lasts for 8 turns in average. Does not stack with itself. Does not affect undead."} } }} }
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[407] then
		local current_attack = wesnoth.get_variable("melee_attack")
		current_attack.damage = current_attack.damage * (1.2 + 0.3 * skills_modified[407])
		current_attack.description = "targeted thurst"
		local percentage = 1.2 + 0.3 * skills_modified[407]
		for i = 1,#current_attack do
			if current_attack[i][1] == "specials" then
				table.insert(current_attack[i][2], { "chance_to_hit", { id="targeted thurst" , name="targeted thurst", description="Targeted thurst: This attack's chance to hit is decreased by 10%, but its damage is increased by " .. tostring(percentage*100) .. "%." , sub=10, apply_to="self", cumulative="yes"  } } )
			end
		end
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[408] then
		local damage = (7 + 2 * (skills_modified[408] + skills_modified[408] ^ 2)) * (1 + magic_damage / 100)
		local current_attack = { name="shadow blast", description="shadow blast", range="spell", icon="attacks/magic-missile.png", damage=damage, number=1, type="arcane", defense_weight=0, {"specials" , { { "chance_to_hit" , { id="spell", name="spell", mana=5, description="Spell: This attack is a spell, so it costs 5 mana and has a 100% chance to hit.", value=100, cumulative="no"  } } , { "drain" , { id="drains", name="drains", description="Drain: This unit drains health from living units, healing itself for half the amount of damage it deals." } } }} }
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[410] then
		local current_attack = wesnoth.get_variable("melee_attack")
		current_attack.damage = current_attack.attacks * (0.5 + 0.2 * skills_modified[410])
		current_attack.defense_weight = 0
		current_attack.description = "utterdark corrosion"
		for i = 1,#current_attack do
			if current_attack[i][1] == "specials" then
				table.insert(current_attack[i][2], { "dummy", { id="uttterdark corrosion" , name="uttterdark corrosion", description="Uttterdark Corrosion: The target will take 8 damage per turn until death, and then it will become a loyal servant."  } } )
			end
		end
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[413] then
		local amount = 8 + skills_modified[413] * 8
		for i=1,#unit do
			if unit[i][1] == "abilities" then
				table.insert(unit[i][2], { "dummy", { id="toxic aura" , name="toxic aura", description="This unit carries a lot of poison around, and can poison a lot of units to take " .. tostring(amount) .. " damage per turn, with average duration 7 turns." } } )
			end
		end
	end
	if skills_modified[414] then
		local current_attack = wesnoth.get_variable("melee_attack")
		current_attack.damage = current_attack.damage * (0.9 + 0.2 * skills_modified[414])
		current_attack.description = "feint"
		attack_weight = 0
		for i = 1,#current_attack do
			if current_attack[i][1] == "specials" then
				table.insert(current_attack[i][2], { "chance_to_hit", { id="feint" , name="feint", name_inactive="feint",  description="Feint: This attack's chance to hit set to 100% once that the opponent misses." , description_incative="Feint: This attack's chance to hit set to 100% once that the opponent misses." , value=100 , apply_to="self", cumulative="no" , { "filter_self" , { {  "filter_wml" , { { "variables" ,  {feint=true }}   }}  }}    }} )
			end
		end
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[415] then
		for i=1,#unit do
			if unit[i][1] == "abilities" then
				if skills_modified[415] < 5 then
					local pattern = "*^F*,*^V*"
					if skills_modified[415] >= 2 then
						pattern = pattern .. ",U*"
					end
					if skills_modified[415] >= 3 then
						pattern = pattern .. ",*M*,*H*"
					end
					if skills_modified[415] >= 4 then
						pattern = pattern .. ",*W*,*D*"
					end
					table.insert(unit[i][2], { "hides", { id="nightstalk" , name="nightstalk", name_inactive="nightstalk", description="Nightstalk: This unit can hide on a variety of terrains at night, the higher the skill is the more terrains it can hide on." , affect_self=true , { "filter_self", {  { "filter_location" , {   terrain=pattern, time_of_dat="chaotic" } }  } } } } )
				end
				if skills_modified[415] >= 5 then
					table.insert(unit[i][2], { "hides", { id="nightstalk" , name="nightstalk", name_inactive="nightstalk", description="Nightstalk: This unit is invisible at night." , affect_self=true , { "filter_self", {  { "filter_location" , {   time_of_dat="chaotic" } }  } } } } )
				end
				if skills_modified[415] > 5 then
					local pattern = "*^F*,*^V*"
					if skills_modified[415] >= 7 then
						pattern = pattern .. ",U*"
					end
					if skills_modified[415] >= 8 then
						pattern = pattern .. ",*M*,*H*"
					end
					if skills_modified[415] >= 9 then
						pattern = pattern .. ",*W*,*D*"
					end
					table.insert(unit[i][2], { "hides", { id="stalk" , name="stalk", name_inactive="stalk", description="Stalk: This unit can hide on a variety of terrains also during twilight, the higher the skill is the more terrains it can hide on." , affect_self=true , { "filter_self", {  { "filter_location" , {   terrain=pattern, time_of_dat="neutral" } }  } } } } )
				end


			end
		end
	end
	if skills_modified[416] then
		local amount = (12 + 2 * (skills_modified[416] + skills_modified[416] ^ 2)) * (1 + magic_damage / 100)
		local current_attack = { name="shadow explosion", description="shadow explosion", range="spell", icon="attacks/noctum.png", damage=amount, number=0, type="arcane", defense_weight=0, {"specials" , { { "chance_to_hit" , { id="spell", name="spell", mana=12, description="Spell: This attack is a spell, so it costs 12 mana and has a 100% chance to hit.", value=100, cumulative="no"  } }}} }
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[417] then
		local current_attack = wesnoth.get_variable("melee_attack")
		current_attack.number = 1
		current_attack.damage = current_attack.damage * (0.8 + 0.2 * skills_modified[417])
		current_attack.defense_weight = 0
		current_attack.description = "poison dance"
		local value = 2 * skills_modified[217]
		for i = 1,#current_attack do
			if current_attack[i][1] == "specials" then
				table.insert(current_attack[i][2], { "attacks", { id="poison dance" , name="poison dance", description="Poison Dance: Deals damage and poisons all adjacent enemies, while they cannot retaliate." , value=0, apply_to="opponent" } } )
			end
		end
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[501] then
		local amount = 6 + skills_modified[501] * 6
		for i=1,#unit do
			if unit[i][1] == "abilities" then
				table.insert(unit[i][2], { "heals", { id="healing" , name="heals", description="This unit combines herbal remedies with magic to heal units more quickly than is normally possible on the battlefield.\n\n A unit cared for by this healer may heal up to " .. tostring(amount) .. " HP per turn, or stop poison from taking effect for that turn." , affect_allies=true , affect_self=true, poison="cured", value=amount, { "affect_adjacent", {  adjacent="n,ne,se,s,sw,nw" } } } } )
			end
		end
	end
	if skills_modified[503] then
		local damage = (10 + 3 * (skills_modified[503] + skills_modified[503] ^ 2)) * (1 + magic_damage / 100)
		local current_attack = { name="lightbeam", description="lightbeam", range="spell", icon="attacks/lightbeam.png", damage=damage, number=1, type="arcane", defense_weight=0, {"specials" , { { "chance_to_hit" , { id="spell", name="spell", mana=6, description="Spell: This attack is a spell, so it costs 6 mana and has a 100% chance to hit.", value=100, cumulative="no"  } } }} }
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[505] then
		local amount = 2 + skills_modified[505] * 2
		for i=1,#unit do
			if unit[i][1] == "abilities" then
				table.insert(unit[i][2], { "resistance", { id="prayer" , name="prayer", description="This unit's magical abilities will increase the resistances of any adjacent unit by " .. tostring(amount) .. "%." , add=value , max_value=80, cumulative=yes, affect_self=true, affect_allies=true, affect_enemies=false, { "affect_adjacent", {  adjacent="n,ne,se,s,sw,nw" } } } } )
			end
		end
	end
	if skills_modified[506] then
		local current_attack = wesnoth.get_variable("melee_attack")
		current_attack.defense_weight = 0
		current_attack.description = "divine fury"
		current_attack.damage = current_attack.damage * 0.8
		current_attack.number = current_attack.number + 1
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[508] then
		local current_attack = { name="divine blast", description="divine blast", range="spell", icon="attacks/faerie-fire.png~CS(10,10,40)", damage=24*(1 + magic_damage / 100), number=1, type="arcane", defense_weight=0, {"specials" , { { "chance_to_hit" , { id="spell", name="spell", mana=11, description="Spell: This attack is a spell, so it costs 11 mana and has a 100% chance to hit.", value=100, cumulative="no"  } } }} }
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[510] then
		local current_attack = { name="searing light", description="searing light", range="spell", icon="attacks/faerie-fire.png~CS(20,-30,-80)", damage=(30 + 6 * skills.skill_710 ^ 2) * (1 + (magic_damage + fire_damage)/100), number=1, type="fire", defense_weight=0, {"specials" , { { "chance_to_hit" , { id="spell", name="spell", mana=12, description="Spell: This attack is a spell, so it costs 12 mana and has a 100% chance to hit.", value=100, cumulative="no"  } } }} }
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[513] then
		local current_attack = wesnoth.get_variable("melee_attack")
		current_attack.attacks = current_attack.attacks * (1 + 0.2 * skills_modified[513])
		current_attack.description = "smite"
		for i = 1,#current_attack do
			if current_attack[i][1] == "specials" then
				table.insert(current_attack[i][2], { "damage", { id="smite" , name="smite", description="Smite: The opponent retaliates 25% less damage." , multiply=0.75 , apply_to="opponent", cumulative="yes"  } } )
			end
		end
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[514] then
		local current_attack = { name="divine wrath", description="divine wrath", range="spell", icon="attacks/fire-blast.png~GS()", damage=(10 + 2 * skills.skill_714 ^ 2) * (1 + magic_damage/100), number=1, type="arcane", defense_weight=0, {"specials" , { { "chance_to_hit" , { id="spell", name="spell", mana=14, description="Spell: This attack is a spell, so it costs 14 mana and has a 100% chance to hit.", value=100, cumulative="no"  } } }} }
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[515] then
		local amount = (12 + skills_modified[515] * 12) * (1 + magic_damage / 100)
		for i=1,#unit do
			if unit[i][1] == "abilities" then
				table.insert(unit[i][2], { "dummy", { id="divine retribution" , name="divine retribution", description="Divine Retribution: This unit is protected by a god. Anything that hits this unit takes " ..  tostring(amount) .. " damage." } } )
			end
		end
	end
	if skills_modified[517] then
		local damage = (12 + 3 * skills_modified[517] ^ 2)*(1 + (magic_damage)/100)
		local current_attack = { name="meteor storm", description="meteor storm", range="spell", icon="attacks/firestorm.png", damage=damage, number=1, type="impact", defense_weight=0, {"specials" , { { "chance_to_hit" , { id="spell", name="spell", mana=12, description="Spell: This attack is a spell, so it costs 12 mana and has a 100% chance to hit.", value=100, cumulative="no" } } }} }
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[603] then
		local damage = (7 + 2 * (skills_modified[603] + skills_modified[603] ^ 2)) * (1 + (magic_damage + cold_mastery) / 100)
		local current_attack = { name="freezing bolt", description="freezing bolt", range="spell", icon="attacks/iceball.png", damage=damage, number=1, type="cold", defense_weight=0, {"specials" , { { "chance_to_hit" , { id="spell", name="spell", mana=5, description="Spell: This attack is a spell, so it costs 5 mana and has a 100% chance to hit.", value=100, cumulative="no" } } , { "slow" , { id="slow", name="slows", description="This attack slows the target until it ends a turn. Slow halves the damage caused by attacks and the movement cost for a slowed unit is doubled. A unit that is slowed will feature a snail icon in its sidebar information when it is selected." } } }} }
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[604] then
		local damage = (11 + 3 * (skills_modified[604] + skills_modified[604] ^ 2)) * (1 + (magic_damage + fire_mastery) / 100)
		local current_attack = { name="inferno", description="inferno", range="spell", icon="attacks/fire-blast.png", damage=damage, number=1, type="fire", defense_weight=0, {"specials" , { { "chance_to_hit" , { id="spell", name="spell", mana=6, description="Spell: This attack is a spell, so it costs 6 mana and has a 100% chance to hit.", value=100, cumulative="no" } } , { "dummy", { id="toxin" , name="toxin", description="Toxin: This attack inflicts a status effect that deals 24 damage per turn and lasts for 8 turns in average. Does not stack with itself. Does not affect undead."} } }} }
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[605] then
		resist_fire = resist_fire + (skills_modified[605] + 1) * (intelligence / 8)
		resist_cold = resist_cold + (skills_modified[605] + 1) * (intelligence / 8)
		resist_arcane = resist_arcane + (skills_modified[605] + 1) * (intelligence / 8)
	end
	if skills_modified[606] then
		local damage = (7 + 2 * (skills_modified[606] + skills_modified[606] ^ 2)) * (1 + (magic_damage + cold_mastery) / 100)
		local current_attack = { name="ice nova", description="ice nova", range="spell", icon="attacks/nova.png", damage=damage, number=1, type="cold", defense_weight=0, {"specials" , { { "chance_to_hit" , { id="spell", name="spell", mana=9, description="Spell: This attack is a spell, so it costs 9 mana and has a 100% chance to hit.", value=100, cumulative="no" } } , { "slow" , { id="slow", name="slows", description="This attack slows the target until it ends a turn. Slow halves the damage caused by attacks and the movement cost for a slowed unit is doubled. A unit that is slowed will feature a snail icon in its sidebar information when it is selected." } } }} }
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[608] then
		local damage = (4 + 3 * (skills_modified[608] + skills_modified[608] ^ 2)) * (1 + (magic_damage + cold_mastery) / 100)
		local current_attack = { name="grimcold blast", description="grimcold blast", range="spell", icon="attacks/iceball.png~CS(-50,0,-50)", damage=damage, number=1, type="cold", defense_weight=0, {"specials" , { { "chance_to_hit" , { id="spell", name="spell", mana=11, description="Spell: This attack is a spell, so it costs 11 mana and has a 100% chance to hit.", value=100, cumulative="no" } } , { "drain" , { id="drains", name="drains", description="Drain: This unit drains health from living units, healing itself for half the amount of damage it deals." } } }} }
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[609] then
		local damage = (12 + 4 * (skills_modified[609] + skills_modified[609] ^ 2)) * (1 + (magic_damage + fire_mastery) / 100)
		local current_attack = { name="fire blast", description="fire blast", range="spell", icon="attacks/fire-blast.png~CS(-10,-20,-50)", damage=damage, number=1, type="fire", defense_weight=0, {"specials" , { { "chance_to_hit" , { id="spell", name="spell", mana=10, description="Spell: This attack is a spell, so it costs 10 mana and has a 100% chance to hit.", value=100, cumulative="no" } } }} }
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[613] then
		local damage = (24 + 2 * (skills_modified[613] + skills_modified[613] ^ 2))*(1 + (magic_damage + fire_mastery)/100)
		local current_attack = { name="meteor storm", description="meteor storm", range="spell", icon="attacks/firestorm.png", damage=damage, number=1, type="fire", defense_weight=0, {"specials" , { { "chance_to_hit" , { id="spell", name="spell", mana=12, description="Spell: This attack is a spell, so it costs 12 mana and has a 100% chance to hit.", value=100, cumulative="no" } } }} }
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[610] then
		for j = 1,#unit do
			if unit[j][1] == "abilities" then
				table.insert(unit[j][2],{ "resistance" , { id="debilitating aura" , name="debilitating aura" , description="debilitating aura: Lowers the resistances of all adjacent enemies by " .. tostring(5 + skills_modified[610]*5) .. "%." , sub=5 + skills_modified[610]*5 , max_value = 80 , apply_to = "fire,cold,arcane,blade,impact,pierce", affect_enemies=true, affect_allies=false, affect_self=false, { "affect_adjacent" , { adjacent="n,ne,se,s,sw,nw"  }} }})
			end
		end	
	end
	if skills_modified[611] then
		local damage = (2 + 3 * (skills_modified[611] + skills_modified[611] ^ 2)) * (1 + (magic_damage + cold_mastery) / 100)
		local current_attack = { name="chilling beam", description="chilling beam", range="spell", icon="attacks/iceball.png", damage=damage, number=1, type="cold", defense_weight=0, {"specials" , { { "chance_to_hit" , { id="spell", name="spell", mana=16 - 2 * skills_modified[611] , description="Spell: This attack is a spell, so it costs " .. 16 - 2 * skills_modified[611] .." mana and has a 100% chance to hit.", value=100, cumulative="no" } } }} }
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[614] then
		for j = 1,#unit do
			if unit[j][1] == "abilities" then
				table.insert(unit[j][2],{ "dummy" , { id="counterspell" , name="counterspell" , description="counterspell: Anybody who attacks this unit with cold, magical or arcane damage will take " .. tostring(skills_modified[614] * 12 * (1 + (magic_damage + cold_mastery) / 100)) .. " cold damage back and will be slowed."}})
			end
		end	
	end
	if skills_modified[703] then
		local current_attack = { name="dark blast", description="dark blast", range="spell", icon="misc/dark-missile.png", damage=18, number=1, type="arcane", defense_weight=0, {"specials" , { { "chance_to_hit" , { id="spell", name="spell", mana=6, description="Spell: This attack is a spell, so it costs 6 mana and has a 100% chance to hit.", value=100, cumulative="no"  } } , { "drain" , { id="drains", name="drains", description="Drain: This unit drains health from living units, healing itself for half the amount of damage it deals." } } }} }
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[705] then
		local damage = (7 + 2 * (skills_modified[705] + skills_modified[705] ^ 2)) * (1 + (magic_damage + cold_mastery) / 100)
		local current_attack = { name="freezing wave", description="freezing wave", range="spell", icon="attacks/iceball.png", damage=damage, number=1, type="cold", defense_weight=0, {"specials" , { { "chance_to_hit" , { id="spell", name="spell", mana=10, description="Spell: This attack is a spell, so it costs 10 mana and has a 100% chance to hit.", value=100, cumulative="no" } } , { "slow" , { id="slow", name="slows", description="This attack slows the target until it ends a turn. Slow halves the damage caused by attacks and the movement cost for a slowed unit is doubled. A unit that is slowed will feature a snail icon in its sidebar information when it is selected." } } }} }
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[707] then
		local damage = (7 + 2 * (skills_modified[707] + skills_modified[707] ^ 2)) * (1 + magic_damage / 100)
		local current_attack = { name="shades of the fallen", description="shades of the fallen", range="spell", icon="attacks/dark-missile.png~GS~CS(20,-100,-100)", damage=damage, number=1, type="arcane", defense_weight=0, {"specials" , { { "chance_to_hit" , { id="spell", name="spell", mana=6, description="Spell: This attack is a spell, so it costs 6 mana and has a 100% chance to hit.", value=100, cumulative="no"  } },{ "dummy", { id="toxin" , name="toxin", description="Toxin: This attack inflicts a status effect that deals 24 damage per turn and lasts for 8 turns in average. Does not stack with itself. Does not affect undead."} } }} }
		unit[#unit + 1] = { "attack" , current_attack}
	end
	if skills_modified[611] then
		local damage = (3 + 3 * (skills_modified[611] + skills_modified[611] ^ 2)) * (1 + magic_damage / 100)
		local current_attack = { name="rabid fumes", description="rabid fumes", range="spell", icon="attacks/bane.png", damage=damage, number=1, type="arcane", defense_weight=0, {"specials" , { { "chance_to_hit" , { id="spell", name="spell", mana=12 , description="Spell: This attack is a spell, so it costs 12 mana and has a 100% chance to hit.", value=100, cumulative="no" } } }} }
		unit[#unit + 1] = { "attack" , current_attack}
	end
	for i=1,#unit do
		--Computing the resistance
		if unit[i][1]=="resistance" then
			if skills_modified[701] then
				unit.alignment = "chaotic"
				for j=1,#unit do
					if unit[j][1] == "status" then
						unit[j][2].not_living = true
					end
				end
				unit[i][2].pierce = unit[i][2].pierce - 5 * skills_modified[701]
				unit[i][2].blade = unit[i][2].blade - 3 * skills_modified[701]
				unit[i][2].cold = unit[i][2].cold - 6 * skills_modified[701]
			end
			if defense ~= 0 then
				-- This is probably the best formula that can be used for calculating the damage decrease from defence in wesnoth, remember it is in combination with absorb
				unit[i][2].pierce = unit[i][2].pierce + 100 * 2 ^ ( -0.02 * defense ^ 0.6) - 100
				unit[i][2].blade = unit[i][2].blade + 100 * 2 ^ ( -0.02 * defense ^ 0.6) - 100
				unit[i][2].impact = unit[i][2].impact + 100 * 2 ^ ( -0.015 * defense ^ 0.5) - 100	-- Maces penetrate armors better
			end
			-- Resistances should be capped. I should have coded it better, using something like resist["fire"] in order to avoid this long code, but I was a noob to lua when I was starting.
			-- Now it's quite hard to change the whole thing, so I am leaving it be, a perfect code is not needed. A pity that lua can't use macros like C, they would make this look better.
			if resist_fire ~= 0 then
				unit[i][2].fire = unit[i][2].fire - resist_fire
				if unit[i][2].fire < 20 then
					unit[i][2].fire = 20
				end
			end
			if resist_cold ~= 0 then
				unit[i][2].cold = unit[i][2].cold - resist_cold
				if unit[i][2].cold < 20 then
					unit[i][2].cold = 20
				end
			end
			if resist_arcane ~= 0 then
				unit[i][2].arcane = unit[i][2].arcane - resist_arcane
				if unit[i][2].arcane < 30 then
					unit[i][2].arcane = 30
				end
			end
			if resist_pierce ~= 0 then
				unit[i][2].pierce = unit[i][2].pierce - resist_pierce
				if unit[i][2].pierce < 20 then
					unit[i][2].pierce = 20
				end
			end
			if resist_blade ~= 0 then
				unit[i][2].blade = unit[i][2].blade - resist_blade
				if unit[i][2].blade < 20 then
					unit[i][2].blade = 20
				end
			end
			if resist_impact ~= 0 then
				unit[i][2].impact = unit[i][2].impact - resist_impact
				if unit[i][2].impact < 20 then
					unit[i][2].impact = 20
				end
			end
		end
	end
	for i=1,#unit do
		if unit[i][1]=="defense" then
			if skills_modified[1] then
				-- The elvish skill makes defences fall into better defence groups, technically redoing the woodland movement_type to woodland_float movement_type in several steps
				if skills_modified[1] == 1 then
					unit[i][2].frozen = 60
					unit[i][2].sand = 60
				end
				if skills_modified[1] == 2 then
					unit[i][2].frozen = 60
					unit[i][2].sand = 60				
					unit[i][2].swamp = 60
					unit[i][2].reef = 60
				end
				if skills_modified[1] == 3 then
					unit[i][2].frozen = 60
					unit[i][2].sand = 60				
					unit[i][2].swamp = 60
					unit[i][2].shallow_water = 60
					unit[i][2].flat = 50
				end
				if skills_modified[1] >= 4 then
					unit[i][2].frozen = 60
					unit[i][2].sand = 60				
					unit[i][2].swamp = 60
					unit[i][2].shallow_water = 60
					unit[i][2].flat = 50
					unit[i][2].reef = 50
				end
				if skills_modified[1] >= 5 then
					unit.moves = unit.moves + 1
				end
			end
			for terrain,value in pairs(unit[i][2]) do
				-- All terrain defences have base values now. Every former number is computed using a different formula, bacause we must avoid letting some defences getting too high,
				-- because a decrease in chance to hit from 80% to 70% means 12.5% less hits taken, while a decrease from 20% to 10% means 50% less hits taken 
				if dexterity < dex_req then
					-- Painful penalty for using items with too high dexterity requirements
					unit[i][2][terrain] = 100
				else
					if value > 70 then
						if dexterity <= 100 then
							unit[i][2][terrain] = 90 - 1.25 * dexterity + 0.01 * dexterity ^ 2 - 0.000025 * dexterity ^ 3 - cth_modifier * 1.5
						else
							unit[i][2][terrain] = 40
						end
					elseif value <= 70 and value >60 then
						if dexterity <= 100 then
							unit[i][2][terrain] = 80 - 1.2 * dexterity + 0.0098 * dexterity ^ 2 - 0.000025 * dexterity ^ 3 - cth_modifier * 1.3
						else
							unit[i][2][terrain] = 33
						end
					elseif value <= 60 and value >50 then
						if dexterity <= 100 then
							unit[i][2][terrain] = 70 - 0.975 * dexterity + 0.0075 * dexterity ^ 2 - 0.00002 * dexterity ^ 3 - cth_modifier
						else
							unit[i][2][terrain] = 27
						end
					elseif value <= 50 and value >40 then
						if dexterity <= 100 then
							unit[i][2][terrain] = 60 - dexterity + 0.0095 * dexterity ^ 2 - 0.00003 * dexterity ^ 3 - cth_modifier * 0.8
						else
							unit[i][2][terrain] = 25
						end
					elseif value <= 40 and value >30 then
						if dexterity <= 100 then
							unit[i][2][terrain] = 50 - 0.85 * dexterity + 0.009 * dexterity ^ 2 - 0.000033 * dexterity ^ 3 - cth_modifier * 0.7
						else
							unit[i][2][terrain] = 23
						end
					elseif value <=30 then
						if dexterity <= 100 then
							unit[i][2][terrain] = 40 - 0.6 * dexterity + 0.007 * dexterity ^ 2 - 0.00003 * dexterity ^ 3 - cth_modifier / 2
						else
							unit[i][2][terrain] = 20
						end
					end
					if skills_modified[409] then
						unit[i][2][terrain] = unit[i][2][terrain] * 0.96 ^ skills_modified[409]
					end
				end
			end
		end
	end
	if strength < str_req then
		-- Devastating penalty for wearing items with too high strength requirements
		unit.max_moves = 1
	end
	if intelligence < int_req then
		-- Painful penalty for wearing items with too high intelligence requirements
		for i = 1,#unit do
			if unit[i][1] == "variables" then
				unit[i][2].mana_regen = 0
			end
		end
	end
	unit = wesnoth.set_graphics(unit)
	unit = wesnoth.display_mana(unit)
	wesnoth.set_variable("melee_attack",nil)	-- It might be fired inside synchronize_choice, and setting up some WML variables and letting them exist might cause OoS
	wesnoth.set_variable("ranged_attack",nil)
	return unit
end

function wesnoth.wml_actions.show_unit_properties()
	-- This function handles the window that shows unit properties. Its subfunctions can also modify it significantly, like when the unit is imported or when the unit advances.
	-- These changes modify the unit code quite deeply, so the unit's code can change wildly, and wesnoth.synchronize_choice cannot return variable arrays. Fortunately, the encryption
	-- that is used for exporting can be used here too.
	local side_number = wesnoth.get_variable("side_number")
	local raw_unit = wesnoth.get_units { canrecruit=true , side=side_number }
	-- Has to be executed on a single machine
	local result = wesnoth.synchronize_choice(
		function()
			-- Sets up some stuff
			local unit = raw_unit[1].__cfg
			local gender = unit.gender
			local name = unit.name
			local new_unit_code
			local dialog = { T.tooltip { id = "tooltip_large" }, T.helptip { id = "tooltip_large" },  maximum_height = 600, maximum_width = 800,
				T.grid { T.row { T.column { T.label { id="title" , definition="title" , label="Character Properties"} }} ,
					T.row { T.column { T.grid { T.row { T.column { T.label { id = "gui_experience_display", label="Experience: 0/0"}},
						T.column { T.button { id = "gui_level_up", label = "Level-up" , tooltip="Opens the advancing window." }}
					}}}},
					T.row { T.column { T.grid { T.row { T.column { T.button { id = "gui_export", label = "Export character" , tooltip="Move the character outside the save file to load it later, in another game. It is recommended to do this if you are leaving a game." }},
						T.column { T.button { id = "gui_import", label = "Import character" , tooltip="Imports a previously exported character." }}
					}}}},
					T.row { T.column { T.scroll_label { id = "gui_message", label = " " } }},
					T.row { T.column { T.button { id = "ok", label = "Back" , tooltip="Return to game." } }}

				}}
			local function preshow()
				local function describe(unit)
					local race, intelligence, strength, dexterity, endurance, total_level, melee_damage, melee_damage_plus, ranged_damage, ranged_damage_plus, melee_precision, ranged_precision, melee_attacks, melee_attacks_plus, ranged_attacks, ranged_attacks_plus, movement, resist_fire, resist_cold, resist_arcane, resist_blade, resist_impact, resist_pierce, skills, specials_melee, specials_ranged, defense, absorb, skills_modified, melee_suck, melee_suck_mana, ranged_suck, ranged_suck_mana, penetrate_fire, penetrate_cold, penetrate_arcane, penetrate_blade, penetrate_impact, penetrate_pierce, magic_damage, melee_type, ranged_type, mana_regen, hitpoints, courage, effects, alignment, class_levels, mana, weapon_type, str_req, int_req, dex_req = wesnoth.get_unit_properties(unit)
					-- Nicely print a description, containing all stuff that might be important I can think of
					local resistances
					local function round(number)
						return (number - number % 1)
					end
					local function edit_to_resist(number)
						return (100 - round(number)) .. "%"
					end
					for i=1,#unit do
						if unit[i][1]=="resistance" then
							resistances = unit[i][2]
						end
						if unit[i][1]=="variables" then
							variables = unit[i][2]
						end
					end
					local spells_names,spells_costs,spells_descriptions,spells_names_ingame,enchantment_cost = wesnoth.setup_spells_descriptions(variables,magic_damage)
					local message = "\nName: " .. unit.name .. "\nRace: " .. race .. "\nAlignment: " .. unit.alignment .. "\nLevel: " ..  total_level .. "\nDefense: " .. defense .. "\n\nAttributes:\nStrength: " .. strength .. "\nDexterity: " .. dexterity .. "\nEndurance: " .. endurance .. "\nIntelligence: " .. intelligence .. "\n\nResistances:\nArcane: " .. edit_to_resist(resistances.arcane) .. "\nFire: " .. edit_to_resist(resistances.fire) .. "\nCold: " .. edit_to_resist(resistances.cold) .. "\nBlade: " .. edit_to_resist(resistances.blade) .. "\nImpact: " .. edit_to_resist(resistances.impact) .. "\nPierce: " .. edit_to_resist(resistances.pierce) .. "\n\nAttacks:" 
					for i=1,#unit do
						if unit[i][1]=="attack" then
							message = message .. "\n" .. unit[i][2].description .. ": " .. round(unit[i][2].damage) .. "-" .. round(unit[i][2].number) .. " " .. unit[i][2].type .. " (" .. unit[i][2].range .. ")"
						end
					end
					message = message .. "\n\nSpells:"
					local found_spells = false
					for i,j in pairs(spells_names_ingame) do
						found_spells = true
						message = message .. "\n" .. j .. ": " .. spells_descriptions[i]
						if spells_costs[i] > 0 then
							message = message .. " (Mana cost: " .. spells_costs[i] .. ")"
						end
						if enchantment_cost[i] > 0 then
							message = message .. " (Mana cost: " .. enchantment_cost[i] .. " per turn)"
						end
					end
					if found_spells == false then
						message = message .. "\nnone"
					end
					return message
				end
				local function export()
					-- Export character - one of the most important features here. It is quite simple, all the important properties are written in the [a] tags in [variables],
					-- so all we need are these codes, that can be written in a comma-separated list, the gender and name (that should be easily changeable by editing the file,
					-- because it does not help anybody to win)
					wesnoth.wml_actions.get_global_variable { namespace="various_world_of_wesnoth", from_global="heroes_vault", to_local="heroes_vault", side = side_number }
					local heroes = wesnoth.get_variable( "heroes_vault" )
					-- We've obtained the data from persistent variables
					if #heroes == 0 then
						heroes = {}
					end
					local variables = helper.get_child(unit, "variables")
					-- The ids in [a] tags are copied into a classical lua list
					local list = {}
					for i = 1,#variables do
						if variables[i][1] == "a" then
							table.insert(list , variables[i][2].id)
						end
					end
					local list_finished = table.concat(list, ',')	-- Nice function that transforms lua lists into comma-separated lists
					list_finished = list_finished .. "," .. tostring(4000000 + unit.experience)	-- Experience is not a part of the [a] list, but should be somehow cryptically saved too
					if wesnoth.sides[side_number].gold < 999999 then
						list_finished = list_finished .. "," .. tostring(3000000 + wesnoth.sides[side_number].gold)	-- Gold is not a part of the [a] list, but should be somehow cryptically saved too
					else
						list_finished = list_finished .. ",3999999"
					end
					table.insert(heroes , { "hero" , { name = unit.name , description = list_finished, gender=unit.gender, appearance=variables.baseframe}})
					-- Export the modified list
					wesnoth.set_variable("heroes_vault" , heroes)
					wesnoth.wml_actions.set_global_variable { namespace="various_world_of_wesnoth", to_global="heroes_vault", from_local="heroes_vault", immediate=true, side = side_number }
					wesnoth.set_variable("heroes_vault",nil)
					wesnoth.message_box("Character exported successfully.")
				end
				local function import()
					-- This is mostly a handler for the wesnoth.import_character function for this GUI, it also checks if a unit was actually imported or the window was cancelled
					local creation,new_name,new_gender = wesnoth.import_character({cannot_cancel=false})
					if creation then
						new_unit_code = creation
						name = new_name
						gender = new_gender
						unit = wesnoth.decrypt_unit(creation,name,gender)
					end
					wesnoth.set_dialog_value(describe(unit), "gui_message" )
					wesnoth.set_dialog_value( "Experience: " .. unit.experience .. "/" .. unit.max_experience .. " " , "gui_experience_display")
					if unit.experience >= unit.max_experience then
						wesnoth.set_dialog_active( true, "gui_level_up")
					else
						wesnoth.set_dialog_active( false, "gui_level_up")
					end	
				end
				local function level_up()
					-- This is mostly a handler for the wesnoth.level-up function for this GUI
					local advanced_unit_code = wesnoth.level_up(unit)
					if advanced_unit_code then	-- The changes might have been declined
						new_unit_code = advanced_unit_code
						unit = wesnoth.decrypt_unit(new_unit_code,name,gender)
					end
					wesnoth.set_dialog_value(describe(unit), "gui_message" )
					wesnoth.set_dialog_value( "Experience: " .. unit.experience .. "/" .. unit.max_experience .. " " , "gui_experience_display")
					if unit.experience >= unit.max_experience then
						wesnoth.set_dialog_active( true, "gui_level_up")
					else
						wesnoth.set_dialog_active( false, "gui_level_up")
					end
				end
				wesnoth.set_dialog_value( "Experience: " .. unit.experience .. "/" .. unit.max_experience .. " " , "gui_experience_display")
				-- Unit can level-up only if its experience is high enough
				if unit.experience < unit.max_experience then
					wesnoth.set_dialog_active( false, "gui_level_up")
				end
				if unit.experience >= unit.max_experience then
					wesnoth.set_dialog_active( true, "gui_level_up")
				end
				wesnoth.set_dialog_callback(level_up, "gui_level_up")
				wesnoth.set_dialog_callback(export, "gui_export")
				wesnoth.set_dialog_callback(import, "gui_import")
				wesnoth.set_dialog_value(describe(unit), "gui_message" )
			end

			local function postshow()
			end

			wesnoth.show_dialog(dialog, preshow, postshow)
			return { code = new_unit_code , name = name, gender = gender}
		end
	)
	-- Interpretation of the result, the value of the result variable is all the other clients know.
	if result.code then
		local decrypted_unit = wesnoth.decrypt_unit(result.code, result.name, result.gender)
		wesnoth.put_unit(decrypted_unit)
	end
end

function wesnoth.decrypt_unit(raw_description,name,gender)
	local side_number = wesnoth.get_variable("side_number")
	-- Very important function, converting comma-separated lists of properties into the [a] tags in initialized unit code, that are by its end completely interpreted by wesnoth.update_stats
	local properties = {}
	-- Basic transformation about the unit is converted into the [a] tags in [variables] and experience
	for j in string.gmatch( raw_description , "[^%s,][^,]*") do
		if tonumber(j) >= 3000000 and tonumber(j) < 4000000 then
			wesnoth.sides[side_number].gold = j - 3000000
		else
			if tonumber(j) >= 4000000 then
				experience = j - 4000000
			else
				table.insert( properties , { "a" , { id=j }} )
			end
		end
	end
	-- Load the properties of the previous unit that have to be kept
	local side_number = wesnoth.get_variable("side_number")
	local raw_unit_import = wesnoth.get_units { canrecruit=true , side=side_number }
	unit_import = raw_unit_import[1].__cfg
	local health_percentage = unit_import.hitpoints / unit_import.max_hitpoints
	local mana_percentage
	local id = unit_import.id
	-- Initialize mana
	for i = 1,#unit_import do
		if unit_import[i][1] == "variables" then
			if unit_import[i][2].mana and unit_import[i][2].max_mana then
				mana_percentage = tonumber(unit_import[i][2].mana) / tonumber(unit_import[i][2].max_mana)
			else
				mana_percentage = 1
			end
		end
	end
	properties.mana = 10 * mana_percentage
	properties.max_mana = 10
	properties.mana_to_pay = 0
	-- Create the unit itself, mostly just for update stats, though
	raw_unit_import = wesnoth.create_unit { type="Hero", id=id, experience=experience, gender=gender, x=unit_import.x, y=unit_import.y, canrecruit=true, side=side_number, name=name, moves=0, attacks_left=0, { "variables" , properties }}
	unit_import = raw_unit_import.__cfg
	unit_import.hitpoints = health_percentage * unit_import.max_hitpoints
	-- Enough data in the variable unit_import to cast wesnoth.update_stats upon it
	unit_import = wesnoth.update_stats(unit_import)
	return unit_import
end

function wesnoth.import_character(cfg)
	-- The GUI where the character chooses a character from the list of exported characters. It is more important than complicated. It will be entirely run in wesnoth.synchronize_choice
	local side_number = wesnoth.get_variable("side_number")
	wesnoth.wml_actions.get_global_variable { namespace="various_world_of_wesnoth", from_global="heroes_vault", to_local="heroes_vault", side = side_number }
	local heroes = wesnoth.get_variable( "heroes_vault" )
	wesnoth.set_variable("heroes_vault",nil)
	-- Data about exported units gathered, time to initialize variables
	local unit_import = nil
	local raw_description
	local gender
	local name
	local did_choose = 0
	if #heroes == 0 then
		-- For the case if the list is empty
		if cfg.cannot_cancel == true then
			wesnoth.message_box("You have no characters exported. Character creation window will be open instead.")
			wesnoth.wml_actions.character_creation()
		else
			wesnoth.message_box("You have no characters exported.")
		end
	else
		local character_chosen	-- Will remain nil until a character is picked
		local import_dialog = { T.tooltip { id = "tooltip_large" }, T.helptip { id = "tooltip_large" },  maximum_height = 600, maximum_width = 800,
			T.grid { T.row { T.column { T.label { id="title" , definition="title" , label="Import Hero"} }} ,
				T.row {	T.column { horizontal_grow = true, T.listbox { id = "gui_character_chosen", 
					T.list_definition { T.row { T.column { horizontal_grow = true,
						T.toggle_panel { tooltip="The list of characters you have exported.", T.grid { T.row {
							T.column { T.image { id = "gui_character_image" } },
		   					T.column { T.grid { 
								T.row { T.column { horizontal_alignment = "left", T.label { id = "gui_character_name" } }},
								T.row { T.column { horizontal_alignment = "left", T.label { id = "gui_character_experience" } }}
							}}
						}}}
					}}}
				}}},
				T.row { T.column { T.grid { T.row {
					T.column { T.button { id = "ok", label = "Import character" , tooltip="Import the selected character replacing your current one. It is advised to export your current character first." } },
					T.column { T.button { id = "cancel", label = "Back", tooltip="Back to game." } }
				}}}}
			}}
		local function import_preshow()
			local function select_character()
				wesnoth.set_dialog_active( true, "ok")
				character_chosen = wesnoth.get_dialog_value "gui_character_chosen"
			end
			for i = 1,#heroes do
				-- The player sees the most important data, name, experience (to distinguish between versions of the same character), and appearance (this one is simply saved, not calculated)
				local experience = 0
				for j in string.gmatch( heroes[i][2].description , "[^%s,][^,]*") do
					if tonumber(j) >= 4000000 then
						experience = j - 4000000
					end
				end
				local appearance
				if not heroes[i][2].appearance then
					-- Some default appearance if something fails
					appearance = "character/base/base.png~BLIT(character/armor/leather.png~CS(-30,-40,-40))~BLIT(character/base/human.png)~BLIT(character/weapons/sword.png)~BLIT(character/armor/shield.png)"
				else
					appearance = heroes[i][2].appearance
				end
				wesnoth.set_dialog_value(appearance , "gui_character_chosen", i , "gui_character_image")
				wesnoth.set_dialog_value(heroes[i][2].name , "gui_character_chosen", i , "gui_character_name")
				wesnoth.set_dialog_value("Experience: " .. experience , "gui_character_chosen", i , "gui_character_experience")
			end
			wesnoth.set_dialog_active( false, "ok")
			wesnoth.set_dialog_callback(select_character, "gui_character_chosen")
		end
		local function import_postshow()
		end
		did_choose = wesnoth.show_dialog(import_dialog,import_preshow,import_postshow)
		local experience = 0
		if did_choose == -1 then
			-- Setting up return values
			raw_description = heroes[character_chosen][2].description
			name = heroes[character_chosen][2].name
			gender = heroes[character_chosen][2].gender
		end		
	end
	return raw_description,name,gender,did_choose
end

function wesnoth.wml_actions.import_character(cfg)
	-- Just to handle the case when import_character is invoked by WML, and has to be handled with synchronize_choice
	local encrypted_unit = wesnoth.synchronize_choice(
		function()
			local code,name,gender,did_choose = wesnoth.import_character(cfg)
			return { code = code , name = name, gender = gender, did_choose = did_choose}
		end
	)
	if encrypted_unit.did_choose == - 1 then
		local unit = wesnoth.decrypt_unit(encrypted_unit.code, encrypted_unit.name, encrypted_unit.gender)
		wesnoth.put_unit(unit)
	else
		if cfg.cannot_cancel == true then
			wesnoth.fire_event("side turn 1")
		end
	end
end
