local helper = wesnoth.require "lua/helper.lua"
local T = helper.set_wml_tag_metatable {}
--local debug_utils = wesnoth.require "~add-ons/World_of_Wesnoth/lua/utils.lua"

function wesnoth.wml_actions.put_item(cfg)
	-- This functions places an item on the ground, adding an event that fires an event that handles the item pick-up.
	local number = cfg.id or helper.wml_error "[put_item] missing required id value"
	local coord_x = cfg.x or helper.wml_error "[put_item] missing required x value"
	local coord_y = cfg.y or helper.wml_error "[put_item] missing required y value"
	local raw_items = wesnoth.get_variable "items" or {}
	local items = {}
	-- Get the container variable that lists items on the ground
	items[2]=raw_items
	items[1]="items"
	local item_code = { "item" , { id = number , x = coord_x , y = coord_y}}
	items[2][#items[2]+1] = item_code
	wesnoth.set_variable("items",items[2])
	-- Added the item to the list, the WML event will check through all items and check if something is there

	-- Now, determining what picture to draw on the ground
	if number <70000  and number >= 60000 then
		local item_code_picked = uniques[number-60000][2]
		item_appearance = item_code_picked.image or helper.wml_error("Item " .. number .. " is missing its on-ground image")
		if item_code_picked.graphics_mod then
			item_appearance = item_appearance .. item_code_picked.graphics_mod
		end
	end
	if number <1000000  and number >= 200000 then
		local base_item = base_items[((number - 200000) % 10000 - (number - 200000) % 100)/100][2] or helper.wml_error("Base item of " .. number .. " is not defined")
		item_appearance = base_item.image or helper.wml_error("Item " .. number " is missing its on-ground image")
		if base_item.graphics_mod then
			item_appearance = item_appearance .. base_item.graphics_mod
		end
		if number > 210100 then
			local prefix = prefixes[(number - 200000 - (number - 200000) % 10000)/10000][2] or helper.wml_error("Prefix o f" .. number .. " is not defined")
			if prefix.graphics_mod then
				item_appearance = item_appearance .. prefix.graphics_mod
			end
		end
		if number % 100 > 0 then
			local suffix = suffixes[ number % 100][2] or helper.wml_error("Suffix of " .. number .. " is not defined")
			if suffix.graphics_mod then
				item_appearance = item_appearance .. suffix.graphics_mod
			end
		end
	end
	-- When the appearance is known, the only thing that remains is to place it on the ground with a nice sparkling halo to make it more visible
	if item_appearance then
		wesnoth.wml_actions.item({ x=coord_x , y=coord_y, image=item_appearance, halo="halo/misc/leadership-flare-1.png:20,halo/misc/leadership-flare-2.png:20,halo/misc/leadership-flare-3.png:20,halo/misc/leadership-flare-4.png:20,halo/misc/leadership-flare-4.png:20,halo/misc/leadership-flare-6.png:20,halo/misc/leadership-flare-7.png:20,halo/misc/leadership-flare-8.png:20,halo/misc/leadership-flare-9.png:20,halo/misc/leadership-flare-10.png:20,halo/misc/leadership-flare-11.png:20,halo/misc/leadership-flare-12.png:20,halo/misc/leadership-flare-13.png:20,misc/blank-hex.png:1000" })
	else
		helper.wml_error("Wrong item code " .. number)
	end
	item_appearance = nil
	-- Adding the event to activate
	wesnoth.wml_actions.event({ name = "moveto" , id="ip" .. coord_x .. coord_y , first_time_only= "no", { "filter", { side="1,2,3,4,5", x=coord_x, y=coord_y, canrecruit=true } }, { "fire_event" , { name="item_pickup", { "primary_unit" , { x=coord_x , y=coord_y }  } } }  })

end

function wesnoth.describe_item(number)
	-- This function reads the number of an item and gets its name, readable list of properties, requirements, appearance and price
	if not number then wesnoth.message("describe_item failed: no number") end
	-- Initialization of variables. They all stack additively from various modifiers, so it can't be just a list of properties found
	local increase_intelligence = 0
	local increase_strength = 0
	local increase_dexterity = 0
	local increase_endurance = 0
	local damage = 0
	local damage_plus = 0
	local precision = 0
	local attacks = 0
	local attacks_plus = 0
	local movement = 0
	local resist_fire = 0
	local resist_cold = 0
	local resist_arcane = 0
	local resist_blade = 0
	local resist_impact = 0
	local resist_pierce = 0
	local skills = {}
	local specials = {}
	local defense = 0
	local to_all_skills = 0
	local suck = 0
	local suck_mana = 0
	local absorb = 0
	local penetrate_fire = 0
	local penetrate_cold = 0
	local penetrate_arcane = 0
	local penetrate_blade = 0
	local penetrate_impact = 0
	local penetrate_pierce = 0
	local magic_damage = 0
	local damage_type = nil
	local mana_regen = 0
	local mana = 0
	local hitpoints = 0
	local courage = false
	local effects = {}
	local description = ""
	local item_type = "violin"
	local item_name = "Kalashnikov"		-- Absurdities needed to detect errors
	local item_appearance = "misc/blank_hex.png"
	local alignment = nil
	local str_req = 0
	local dex_req = 0
	local int_req = 0
	local level_req = 1
	local sell_price = 1
	local buy_price = 1
	local function add_item_modifiers(item_code)
		-- Function that gets access to a WML variable holding the properties of a particular item, suffix, prefix or something, and adds it to the sum of attributes
		if item_code.sort then
			item_type = item_code.sort
		end
		if item_code.defense then
			defense = defense + item_code.defense
		end
		if item_code.courage then
			courage=true
		end
		if item_code.movement then
			movement = movement + item_code.movement
		end
		if item_code.alignment then
			alignment = item_code.alignment
		end
		if item_code.resist_fire then
			resist_fire = resist_fire + item_code.resist_fire
		end
		if item_code.resist_cold then
			resist_cold = resist_cold + item_code.resist_cold
		end
		if item_code.resist_arcane then
			resist_arcane = resist_arcane + item_code.resist_arcane
		end
		if item_code.resist_pierce then
			resist_pierce = resist_pierce + item_code.resist_pierce
		end
		if item_code.resist_blade then
			resist_blade = resist_blade + item_code.resist_blade
		end
		if item_code.resist_impact then
			resist_impact = resist_impact + item_code.resist_impact
		end
		if item_code.increase_strength then
			increase_strength = increase_strength + item_code.increase_strength
		end
		if item_code.increase_dexterity then
			increase_dexterity = increase_dexterity + item_code.increase_dexterity
		end
		if item_code.increase_endurance then
			increase_endurance = increase_endurance + item_code.increase_endurance
		end
		if item_code.increase_intelligence then
			increase_intelligence = increase_intelligence + item_code.increase_intelligence
		end
		if item_code.damage then
			damage = damage + item_code.damage
		end
		if item_code.damage_plus then
			damage_plus = damage_plus + item_code.damage_plus
		end
		if item_code.attacks then
			attacks = attacks + item_code.attacks
		end
		if item_code.attacks_plus then
			attacks_plus = attacks_plus + item_code.attacks_plus
		end
		if item_code.precision then
			precision = precision + item_code.precision
		end
		if item_code.suck then
			suck = suck + item_code.suck
		end
		if item_code.suck_mana then
			suck_mana = suck_mana + item_code.suck_mana
		end
		if item_code.damage_type then
			damage_type = item_code.damage_type
		end
		if item_code.mana_regen then
			mana_regen = mana_regen + item_code.mana_regen
		end
		if item_code.mana then
			mana = mana + item_code.mana
		end
		if item_code.magic then
			magic_damage = magic_damage + item_code.magic
		end
		if item_code.hitpoints then
			hitpoints = hitpoints + item_code.hitpoints
		end
		if item_code.to_all_skills then
			to_all_skills = to_all_skills + item_code.to_all_skills
		end
		if item_code.absorb then
			absorb = absorb + item_code.absorb
		end
		if item_code.penetrate_fire then
			penetrate_fire = penetrate_fire + item_code.penetrate_fire
		end
		if item_code.penetrate_cold then
			penetrate_cold = penetrate_cold + item_code.penetrate_cold
		end
		if item_code.penetrate_arcane then
			penetrate_arcane = penetrate_arcane + item_code.penetrate_arcane
		end
		if item_code.penetrate_blade then
			penetrate_blade = penetrate_blade + item_code.penetrate_blade
		end
		if item_code.penetrate_pierce then
			penetrate_pierce = penetrate_pierce + item_code.penetrate_pierce
		end
		if item_code.penetrate_impact then
			penetrate_impact = penetrate_impact + item_code.penetrate_impact
		end
		for aim_i=1,#item_code do
			-- Subtags have to be handled with care
			if item_code[aim_i][1] == "effect" then
				table.insert(effects,item_code[aim_i])
			end
			if item_code[aim_i][1] == "skills" then
				for j,k in pairs(item_code[aim_i][2]) do
					if skills[j] then
						skills[j] = skills[j] + k
					else
						skills[j] = k
					end
				end
			end
			if item_code[aim_i][1] == "specials" then
				table.insert(specials,item_code[aim_i][2])
			end
		end
		-- And also finding the maximum of the main inconveniences - requirements
		if item_code.level then
			if item_code.level > level_req then
				level_req = item_code.level
			end
		end
		if item_code.required_strength then
			if item_code.required_strength > str_req then
				str_req = item_code.required_strength
			end
		end
		if item_code.required_dexterity then
			if item_code.required_dexterity > dex_req then
				dex_req = item_code.required_dexterity
			end
		end
		if item_code.required_intelligence then
			if item_code.required_intelligence > int_req then
				int_req = item_code.required_intelligence
			end
		end
	end
	if number <70000  and number >= 60000 then
		-- For unique items, getting the properties is somewhat straightforward, everything is written on one place
		local item_code_picked = uniques[number-60000][2]
		item_name = item_code_picked.name
		sell_price = ((item_code_picked.price - item_code_picked.price % 10) / 10)	-- The price of uniques is not computed automatically, and the sell price is always 10-times less than the buy price
		buy_price = item_code_picked.price
		add_item_modifiers(item_code_picked)
		item_appearance = item_code_picked.image or helper.wml_error("Item" .. number .. "is missing its on-ground image")
		if item_code_picked.graphics_mod then
			item_appearance = item_appearance .. item_code_picked.graphics_mod
		end
	end
	if number <200000  and number >= 100000 then
		--will be used for crafted items
	end
	if number <1000000  and number >= 200000 then
		-- For magical items, all properties from base item, prefix and suffix have to be summed, so all properties are handled separately
		local base_item = base_items[((number - 200000) % 10000 - (number - 200000) % 100)/100][2] or helper.wml_error("Base item of" .. number .. "is not defined")
		item_appearance = base_item.image or helper.wml_error("Item" .. number "is missing its on-ground image")
		add_item_modifiers(base_item)
		item_name = base_item.name	-- The name will have to be generated as well
		sell_price = base_item.level + 4
		buy_price = base_item.level ^ 1.8 - ((base_item.level ^ 1.8) % 1) + 4	-- These polynomic functions are designed to generate the prices in the way that looks good to me
		if base_item.graphics_mod then
			item_appearance = item_appearance .. base_item.graphics_mod
		end
		if number > 210100 then
			local prefix = prefixes[(number - 200000 - (number - 200000) % 10000)/10000][2] or helper.wml_error("Prefix of" .. number .. "is not defined")
			item_name = prefix.name .. " " .. item_name
			sell_price = sell_price + prefix.level + 4
			-- This formula serves to make sure that combinations of high-level modifiers on high-level items will cost a lot to buy, but not much to sell, as they can be useless anyway
			buy_price = ((buy_price * (prefix.level + 4)) ^ 0.8) - ((buy_price * (prefix.level + 4)) ^ 0.8) % 1
			add_item_modifiers(prefix)
			if prefix.graphics_mod then
				item_appearance = item_appearance .. prefix.graphics_mod
			end
		end
		if number % 100 > 0 then
			local suffix = suffixes[ number % 100][2] or helper.wml_error("Suffix of" .. number .. "is not defined")
			item_name = item_name .. " " .. suffix.name
			sell_price = sell_price + suffix.level + 4
			buy_price = ((buy_price * (suffix.level + 4)) ^ 0.8) - ((buy_price * (suffix.level + 4)) ^ 0.8) % 1
			add_item_modifiers(suffix)
			if suffix.graphics_mod then
				item_appearance = item_appearance .. suffix.graphics_mod
			end
		end
	end
	-- Check for absurdities resulting from errors in execution
	if item_type == "violin" then
		helper.wml_error("describing an item missing its sort (" .. number .. ")")
	end
	if item_name == "Kalashnikov" then
		helper.wml_error("item name identification went wrong (" .. number .. ")")
	else
		local local_item_name = tostring(item_name)
		item_name = local_item_name	
	end
	-- And write the description. Item type is marked first, then followed by requirements, price, and description of effects it possesses.
	if item_type == "melee" or item_type == "wand" then
		item_type = "melee weapon"
	end
	if item_type == "ranged" then
		item_type = "ranged weapon"
	end
	description = "Item type: " .. item_type
	if level_req ~= 1 then
		description = description .. "\n" .. "requires level: " .. tostring(level_req)
	end
	if str_req ~= 0 then
		description = description .. "\n" .. "requires strength: " .. tostring(str_req)
	end
	if dex_req ~= 0 then
		description = description .. "\n" .. "requires dexterity: " .. tostring(dex_req)
	end
	if int_req ~= 0 then
		description = description .. "\n" .. "requires intelligence: " .. tostring(int_req)
	end
	description = description .. "\nSell price: " .. tostring(sell_price) .. " gold"
	if to_all_skills ~= 0 then
		description = description .. "\n" .. to_all_skills .. " to All Skills"
	end
	if damage_type then
		description = description .. "\nDamage Type Set to " .. damage_type
	end
	for j,k in pairs(skills) do
		-- Skills have to be found on the list of skills
		local skill_name =  "Skill name of " .. tostring(k) .. " not found!"
		for i = 1,#classes[(j - j % 100) / 100][2] do
			if classes[(j - j % 100) / 100][2][i][1] == "skill" then
				if j % 100 == classes[(j - j % 100) / 100][2][i][2].number then
					skill_name = classes[(j - j % 100) / 100][2][i][2].name
				end
			end
		end
		description = description .. "\n" .. tostring(k) .. " to " .. skill_name
	end
	if specials[1] then
		-- The description of specials is taken from the WML description
		for j,k in pairs(specials[1]) do
			if k then
				if k[2].name then
					description = description .. "\n" .. k[2].name
				end
			end
		end
	end
	for j,k in pairs(effects) do
		-- [effect] tags used will have to have their own description, or they will be silent
		if k.description then
			description = description .. k.description
		end
	end
	if alignment then
		description = description .. "\nChanges alignment to " .. alignment
	end
	if defense ~= 0 then
		description = description .. "\n" .. defense .. " to Defense"
	end
	if absorb ~= 0 then
		description = description .. "\nAbsorbs " .. absorb .. " damage"
	end
	if damage ~=0 then
		description = description .. "\n" .. damage .. "% More Damage"
	end
	if magic_damage ~=0 then
		description = description .. "\n+" .. magic_damage .. "% to Spells' Damage"
	end
	if attacks ~=0 then
		description = description .. "\n" .. attacks .. "% More Attacks"
	end
	if damage_plus ~=0 then
		description = description .. "\n" .. damage_plus .. " More Damage"
	end
	if attacks_plus ~=0 then
		description = description .. "\n" .. attacks_plus .. " More Attacks"
	end
	if suck ~=0 then
		description = description .. "\nSucks " .. suck .. " Life"
	end
	if suck_mana ~=0 then
		description = description .. "\nSucks " .. suck_mana .. " Mana"
	end
	if penetrate_impact ~=0 then
		description = description .. "\n" .. penetrate_impact .. "% to Impact penetrateance"
	end
	if penetrate_blade ~=0 then
		description = description .. "\n" .. penetrate_blade .. "% to Blade penetrateance"
	end
	if penetrate_pierce ~=0 then
		description = description .. "\n" .. penetrate_pierce .. "% to Pierce penetrateance"
	end
	if penetrate_arcane ~=0 then
		description = description .. "\n" .. penetrate_arcane .. "% to Arcane penetrateance"
	end
	if penetrate_fire ~=0 then
		description = description .. "\n" .. penetrate_fire .. "% to Fire penetrateance"
	end
	if penetrate_cold ~=0 then
		description = description .. "\n" .. penetrate_cold .. "% to Cold penetrateance"
	end
	if precision ~=0 then
		description = description .. "\n" .. precision .. " Better Chance to Hit"
	end
	if increase_endurance ~=0 then
		description = description .. "\n" .. increase_endurance .. " More Endurance"
	end
	if increase_intelligence ~=0 then
		description = description .. "\n" .. increase_intelligence .. " More Intelligence"
	end
	if increase_strength ~=0 then
		description = description .. "\n" .. increase_strength .. " More Strength"
	end
	if increase_dexterity ~=0 then
		description = description .. "\n" .. increase_dexterity .. " More Dexterity"
	end
	if movement ~=0 then
		description = description .. "\n" .. movement .. " More Movement Points"
	end
	if mana_regen ~=0 then
		description = description .. "\n" .. mana_regen .. " More Mana Regenerated"
	end
	if mana ~=0 then
		description = description .. "\n" .. mana .. " More Mana"
	end
	if hitpoints ~=0 then
		description = description .. "\n" .. hitpoints .. " More Life"
	end
	if courage then
		description = description .. "\nCourage"
	end
	if resist_impact ~=0 then
		description = description .. "\n" .. resist_impact .. "% to Impact Resistance"
	end
	if resist_blade ~=0 then
		description = description .. "\n" .. resist_blade .. "% to Blade Resistance"
	end
	if resist_pierce ~=0 then
		description = description .. "\n" .. resist_pierce .. "% to Pierce Resistance"
	end
	if resist_arcane ~=0 then
		description = description .. "\n" .. resist_arcane .. "% to Arcane Resistance"
	end
	if resist_fire ~=0 then
		description = description .. "\n" .. resist_fire .. "% to Fire Resistance"
	end
	if resist_cold ~=0 then
		description = description .. "\n" .. resist_cold .. "% to Cold Resistance"
	end
	return description,item_name,item_type,item_appearance,level_req,str_req,dex_req,int_req,sell_price,buy_price
end

function wesnoth.wml_actions.describe_item(cfg)
	-- This function allows the usage of describe_item with WML, using a pretty and comfortable WML tag
	local number = cfg.number or helper.wml_error("[describe_item] missing the required number attribute")
	local description,item_name,item_type,item_appearance = wesnoth.describe_item(number)
	wesnoth.set_variable("description",description)
	wesnoth.set_variable("item_name",item_name)
	wesnoth.set_variable("item_appearance",item_appearance)
end

function wesnoth.wml_actions.view_inventory(cfg)
	-- This function allows inventory manipulation.
	local items_equipped = {}
	local items_inventory = {}
	local other_things = {}
	local properties = {}
	local side_number = wesnoth.get_variable("side_number")
	local raw_unit = wesnoth.get_units { canrecruit=true , side=side_number }
	local unit = raw_unit[1].__cfg
	local race, intelligence, strength, dexterity, endurance, total_level = wesnoth.get_unit_properties(unit)
	for i=1,#unit do
		if unit[i][1]=="variables" then
			properties = unit[i][2]
		end
	end
	-- Reading all properties of the unit, saving the unimportant ones into the other_things variable, separating the items into two variables, one represents the contents of inventory, other one represents
	-- items that are currently equipped
	for i=1,#properties do
		if properties[i][1] == "a" then
			if properties[i][2].id < 60000 then
				table.insert(other_things,properties[i][2].id)
			end
			if properties[i][2].id <70000  and properties[i][2].id >= 60000 then
				table.insert(items_equipped,properties[i][2].id)
			end
			if properties[i][2].id <200000  and properties[i][2].id >= 100000 then
				table.insert(items_equipped,properties[i][2].id)
			end
			if properties[i][2].id <1000000  and properties[i][2].id >= 200000 then
				table.insert(items_equipped,properties[i][2].id)
			end

			if properties[i][2].id <1070000  and properties[i][2].id >= 1060000 then
				table.insert(items_inventory,properties[i][2].id-1000000)
			end
			if properties[i][2].id <1200000  and properties[i][2].id >= 1100000 then
				table.insert(items_inventory,properties[i][2].id-1000000)
			end
			if properties[i][2].id <2000000  and properties[i][2].id >= 1200000 then
				table.insert(items_inventory,properties[i][2].id-1000000)
			end
		end
	end
	local result = wesnoth.synchronize_choice(
		function()
			-- New variables that will be computed only on one client, some of them are copies of original variables
			local items_inventory_working = wesnoth.deepcopy(items_inventory)
			local items_equipped_working = wesnoth.deepcopy(items_equipped)
			local items_equipped_working_description = {}
			local items_inventory_working_description = {}
			local items_equipped_working_type = {}
			local items_inventory_working_type = {}
			local to_drop = {}
			local last_equipped = 1
			local last_inventory = 1
			local dialog = { T.tooltip { id = "tooltip_large" }, T.helptip { id = "tooltip_large" },  maximum_height = 600, --maximum_width = 800,
				T.grid { T.row { T.column { T.label { id="title" , definition="title" , label="Inventory"} }} ,
					T.row { T.column { T.grid { T.row { T.column {T.button { id = "gui_unequip", label = "Unequip" , tooltip="Unequip the selected item."}},
							T.column {T.button { id = "gui_equip_left", label = "Equip as secondary" , tooltip="Equip the selected item, replacing the second item of that sort the character has equipped."}},
							T.column { T.grid { T.row { 
								T.column { T.button { id = "gui_equip", label = "Equip" , tooltip="Equip the selected item."}},
								T.column { T.button { id = "gui_drop", label = "Drop" , tooltip="Drop the selected item from inventory on the ground."}}
							}}}
						},
						T.row { T.column { horizontal_grow = true, T.listbox { id = "gui_equipped_chosen", 
								T.list_definition { T.row { T.column { horizontal_grow = true,
									T.toggle_panel { tooltip="The list of currently equipped items.", T.grid { T.row {
										T.column { T.image { id = "gui_item_icon" } },
					   					T.column { horizontal_alignment = "left", T.label { id = "gui_item_name" } }
									}}}
								}}}
							}},
							T.column {T.label { id = "gui_description_equipped", label = " " , tooltip="Description of the selected item."}},
							T.column { horizontal_grow = true, T.listbox { id = "gui_inventory_chosen", 
								T.list_definition { T.row { T.column { horizontal_grow = true,
									T.toggle_panel { tooltip="The list of items in inventory.", T.grid { T.row {
										T.column { T.image { id = "gui_item_icon" } },
					   					T.column { horizontal_alignment = "left", T.label { id = "gui_item_name" } }

									}}}
								}}}
							}}
						}
					}}},
					T.row { T.column { T.grid { T.row {
						T.column { T.button { id = "ok", label = "Perform changes" , tooltip="Do all the changes you selected." } },
						T.column { T.button { id = "cancel", label = "Discard changes", tooltip="Undo all selected changes." } }
					}}}}

				}}

-- 				T.column {T.button { id = "gui_crafting", label = "Crafting..." , tooltip="Enter the menu for crafing items."}},

			local function preshow()
				local function select_inventory()
					-- Clearing all slots where something can appear
					for i=1,9 do
						wesnoth.set_dialog_value("                               ", "gui_equipped_chosen", i, "gui_item_name")
						wesnoth.set_dialog_value( "misc/blank-hex.png", "gui_equipped_chosen", i, "gui_item_icon")
					end
					for i=1,(#items_inventory_working + #items_equipped_working) do
						wesnoth.set_dialog_value("                               ", "gui_inventory_chosen", i, "gui_item_name")
						wesnoth.set_dialog_value("misc/blank-hex.png", "gui_inventory_chosen", i, "gui_item_icon")
					end
					-- Fill all slots with items, saving descriptions to be read later
					for i=1,#items_equipped_working do
						local description,item_name,item_type,item_appearance = wesnoth.describe_item(items_equipped_working[i])
						wesnoth.set_dialog_value( item_name , "gui_equipped_chosen", i, "gui_item_name")
						wesnoth.set_dialog_value( item_appearance , "gui_equipped_chosen", i, "gui_item_icon")
						items_equipped_working_description[i] = description
						items_equipped_working_type[i] = item_type
					end
					for i=1,#items_inventory_working do
						local description,item_name,item_type,item_appearance = wesnoth.describe_item(items_inventory_working[i])
						wesnoth.set_dialog_value( item_name , "gui_inventory_chosen", i, "gui_item_name")
						wesnoth.set_dialog_value( item_appearance , "gui_inventory_chosen", i, "gui_item_icon")
						items_inventory_working_description[i] = description
						items_inventory_working_type[i] = item_type
					end
					-- Show the description of the item that is picked
					local preshow_equipped_selected = wesnoth.get_dialog_value "gui_equipped_chosen"
					local preshow_inventory_selected = wesnoth.get_dialog_value "gui_inventory_chosen"
					if items_inventory_working_description[preshow_inventory_selected] then
						wesnoth.set_dialog_value("--------------------------------------\n" .. items_inventory_working_description[preshow_inventory_selected] , "gui_description_equipped")
					else
						wesnoth.set_dialog_value( " " , "gui_description_equipped")
					end
					-- Allow / disallow equipping as secondary according to the item type
					wesnoth.set_dialog_active( false, "gui_equip_left")
					wesnoth.set_dialog_active( true, "gui_equip")
					if items_inventory_working_type[preshow_inventory_selected] == "melee weapon" or items_inventory_working_type[preshow_inventory_selected] == "shield" or items_inventory_working_type[preshow_inventory_selected] == "ring" then
						wesnoth.set_dialog_active( true, "gui_equip_left")
					end
					if items_inventory_working_type[preshow_inventory_selected] == "shield" then
						wesnoth.set_dialog_active( false, "gui_equip")
					end
				end
				local function select_equipped()
					-- Does basically the same as the thing above, just describes another item, could be wrote using the same function
					for i=1,9 do
						wesnoth.set_dialog_value("                                                                                           ", "gui_equipped_chosen", i, "gui_item_name")
						wesnoth.set_dialog_value( "misc/blank-hex.png", "gui_equipped_chosen", i, "gui_item_icon")
					end
					for i=1,(#items_inventory_working + #items_equipped_working) do
						wesnoth.set_dialog_value("                                                                                           ", "gui_inventory_chosen", i, "gui_item_name")
						wesnoth.set_dialog_value("misc/blank-hex.png", "gui_inventory_chosen", i, "gui_item_icon")
					end
					for i=1,#items_equipped_working do
						local description,item_name,item_type,item_appearance = wesnoth.describe_item(items_equipped_working[i])
						wesnoth.set_dialog_value( item_name , "gui_equipped_chosen", i, "gui_item_name")
						wesnoth.set_dialog_value( item_appearance , "gui_equipped_chosen", i, "gui_item_icon")
						items_equipped_working_description[i] = description
						items_equipped_working_type[i] = item_type
					end
					for i=1,#items_inventory_working do
						local description,item_name,item_type,item_appearance = wesnoth.describe_item(items_inventory_working[i])
						wesnoth.set_dialog_value( item_name , "gui_inventory_chosen", i, "gui_item_name")
						wesnoth.set_dialog_value( item_appearance , "gui_inventory_chosen", i, "gui_item_icon")
						items_inventory_working_description[i] = description
						items_inventory_working_type[i] = item_type
					end
					local preshow_equipped_selected = wesnoth.get_dialog_value "gui_equipped_chosen"
					local preshow_inventory_selected = wesnoth.get_dialog_value "gui_inventory_chosen"
					if items_equipped_working_description[preshow_equipped_selected] then
						wesnoth.set_dialog_value( "--------------------------------------\n" .. items_equipped_working_description[preshow_equipped_selected] .. "\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n" , "gui_description_equipped")
					else
						wesnoth.set_dialog_value( " " , "gui_description_equipped")
					end
				end
				local function make_shield_last()
					-- Shield should always be in the second weapon slot. If somebody equips a shield and does not have a weapon in his other hand, it will end up as first weapon,
					-- and will cause trouble - therefore, it has to be moved to be behind the weapon on the list
					local shield_number = 0
					for i = 1,#items_equipped_working do
						if items_equipped_working[i] then
							local description,item_name,item_type = wesnoth.describe_item(items_equipped_working[i])
							if item_type == "shield" then
								shield_number = items_equipped_working[i]
								table.remove(items_equipped_working,i)
							end
						end
					end
					if shield_number ~= 0 then
						items_equipped_working[#items_equipped_working+1] = shield_number
					end
				end
				local function equip()
					-- When equipping an item, we only check if the level requirement is met, requirements for other attributes may be reached from stat increases,
					-- possibly somethinng like the strength bug known in Diablo II - it's better if too high attribute requirements end up in painful penalties instead
					local preshow_inventory_selected = wesnoth.get_dialog_value "gui_inventory_chosen"
					if items_inventory_working[preshow_inventory_selected] then	-- We need to check if an item is selected
						local description2,item_name2,item_type2,item_appearance,level_req = wesnoth.describe_item(items_inventory_working[preshow_inventory_selected])
						if level_req <= total_level then
							local has_it = false
							for i=1,#items_equipped_working do
								if has_it == false then
									local description,item_name,item_type = wesnoth.describe_item(items_equipped_working[i])
									if item_type == item_type2 then
										items_inventory_working[#items_inventory_working+1] = items_equipped_working[i]
										items_equipped_working[i]=items_inventory_working[preshow_inventory_selected]
										table.remove(items_inventory_working,preshow_inventory_selected)
										has_it = true
									end
								end
							end
							if has_it == false then
								items_equipped_working[#items_equipped_working+1] = items_inventory_working[preshow_inventory_selected]
								table.remove(items_inventory_working,preshow_inventory_selected)
							end
							make_shield_last()
							select_equipped()
						else
							wesnoth.message_box("You don't meet the level requirements for this item, so you cannot equip it.")
						end
					end
				end
				local function equip_left()
					-- This is to equip shields, rings and weapons behind the first weapon/ring, possibly replacing the item that is currently there, so we have to find the first item,
					-- mark that something was found, and then replace the second one, or just append it to the end if none is found
					local preshow_inventory_selected = wesnoth.get_dialog_value "gui_inventory_chosen"
					if items_inventory_working[preshow_inventory_selected] then
						local description2,item_name2,item_type2,item_appearance,level_req = wesnoth.describe_item(items_inventory_working[preshow_inventory_selected])
						if level_req <= total_level then
							local has_it = false
							local seen_first = false
							for i=1,#items_equipped_working do
								if has_it == false and seen_first == true then
									local description,item_name,item_type = wesnoth.describe_item(items_equipped_working[i])
									if item_type == "shield" then item_type = "melee weapon" end
									if item_type2 == "shield" then item_type2 = "melee weapon" end
									if item_type == item_type2 then
										items_inventory_working[#items_inventory_working+1] = items_equipped_working[i]
										items_equipped_working[i]=items_inventory_working[preshow_inventory_selected]
										table.remove(items_inventory_working,preshow_inventory_selected)
										has_it = true
									end
								end
								if seen_first == false then
									local description,item_name,item_type = wesnoth.describe_item(items_equipped_working[i])
									if item_type == "shield" then item_type = "melee weapon" end
									if item_type2 == "shield" then item_type2 = "melee weapon" end
									if item_type == item_type2 then
										seen_first = true
									end
								end
							end
							if has_it == false then
								items_equipped_working[#items_equipped_working+1] = items_inventory_working[preshow_inventory_selected]
								table.remove(items_inventory_working,preshow_inventory_selected)
							end
							make_shield_last()
							select_equipped()
						else
							wesnoth.message_box("You don't meet the level requirements for this item, so you cannot equip it.")
						end
					end
				end
				local function unequip()
					-- This one is simpler, no replacing
					local preshow_equipped_selected = wesnoth.get_dialog_value "gui_equipped_chosen"
					if items_equipped_working[preshow_equipped_selected] then
						items_inventory_working[#items_inventory_working+1] = items_equipped_working[preshow_equipped_selected]
						table.remove(items_equipped_working,preshow_equipped_selected)
						select_inventory()
					end
				end
				local function drop()
					-- Basically like unequip, just putting them on another list
					local preshow_inventory_selected = wesnoth.get_dialog_value "gui_inventory_chosen"
					if items_inventory_working[preshow_inventory_selected] then
						to_drop[#to_drop+1] = items_inventory_working[preshow_inventory_selected]
						table.remove(items_inventory_working,preshow_inventory_selected)
						select_inventory()
					end
				end
				-- Setting up callbacks and initialization
				wesnoth.set_dialog_callback(select_equipped, "gui_equipped_chosen")
				wesnoth.set_dialog_callback(select_inventory, "gui_inventory_chosen")
				wesnoth.set_dialog_callback(equip, "gui_equip")
				wesnoth.set_dialog_callback(equip_left, "gui_equip_left")
				wesnoth.set_dialog_callback(unequip, "gui_unequip")
				wesnoth.set_dialog_callback(drop, "gui_drop")
--				wesnoth.set_dialog_callback(crafting, "gui_crafting")
				select_equipped()
			end

			local function postshow()
			end

			local choice_done = wesnoth.show_dialog(dialog, preshow, postshow)
			-- The lists have to be transformed into comma separated lists in strings if they have to be passed through wesnoth.synchronize_choice
			return {choice_done = choice_done, inventory = table.concat(items_inventory_working,",") , equipped = table.concat(items_equipped_working,",") , to_drop = table.concat(to_drop,",")}
		end
	)
	-- Get the list of items back
	items_inventory = {}
	items_equipped = {}
	local to_drop = {}
	for j in string.gmatch( result.inventory , "[^%s,][^,]*") do
		items_inventory[#items_inventory+1] = tonumber(j)
	end
	for j in string.gmatch( result.equipped , "[^%s,][^,]*") do
		items_equipped[#items_equipped+1] = tonumber(j)
	end
	for j in string.gmatch( result.to_drop , "[^%s,][^,]*") do
		to_drop[#to_drop+1] = tonumber(j)
	end
	if result.choice_done == -1 then
		local j = 1
		local problem_message = ""
		-- Recreating the list of [a] tag describing the properties
		while j <= #properties do
			if properties[j][1] == "a" then
				table.remove(properties,j)
			else
				j = j + 1
			end
		end
		for i=1,#other_things do
			properties[#properties+1] = { "a" , {id = other_things[i] } }
		end
		-- Check for requirements, consequences of unmet requirements are computed automatically, but we have to write some warnings.
		for i=1,#items_equipped do
			properties[#properties+1] = { "a" , {id = items_equipped[i] } }
		end
		for i=1,#items_inventory do
			properties[#properties+1] = { "a" , {id = items_inventory[i] + 1000000} }
		end
		for i=1,#unit do
			if unit[i][1]=="variables" then
				unit[i][2] = properties
			end
		end
		-- Drop some items on the ground
		for i=1,#to_drop do
			wesnoth.wml_actions.put_item{ x=unit.x, y=unit.y, id=to_drop[i]}
		end
		unit = wesnoth.update_stats(unit)
		wesnoth.put_unit(unit)
		-- Check for requirements, consequences of unmet requirements are computed automatically, but we have to write some warnings.
		race, intelligence, strength, dexterity, endurance = wesnoth.get_unit_properties(unit)
		for i=1,#items_equipped do
			properties[#properties+1] = { "a" , {id = items_equipped[i] } }
			local description,item_name,item_type,item_appearance,level_req,str_req,dex_req,int_req = wesnoth.describe_item(items_equipped[i])
			if str_req > strength then
				problem_message = item_name .. " is too heavy to carry: strength required is " .. tostring(str_req) .. ", but the character's strength is only " .. tostring(strength) .. ". The character will be able to move only 1 hex per turn.\n"
			end
			if int_req > intelligence then
				problem_message = problem_message .. item_name .. " is too complex to use: intelligence required is " .. tostring(int_req) .. ", but the character's intelligence is only " .. tostring(intelligence) .. ". The character will have no mana regeneration.\n"
			end
			if dex_req > dexterity then
				problem_message = problem_message .. item_name .. " is too hard to manipulate with: dexterity required is " .. tostring(dex_req) .. ", but the character's dexterity is only " .. tostring(dexterity) .. ". The character will not be able to defend.\n"
			end
		end
		-- Show the previously computed problems with requirements
		if problem_message ~= "" then
			wesnoth.message_box(problem_message)
		end
	end

end

function wesnoth.choose_item(var,base_item,level)
	-- This function is passed a variable representing an item list (can be base items, uniques or affixes,...), and it chooses a random item from the list, taking into consideration only the items whose
	-- level is not higher than the level argument, and respecting the chances to drop mentioned. The base_item argument is another filter, that will ignore all items of wrong type (to prevent awkward prefixes
	-- and suffixes from appearing on randomised items).
	local list = ""
	local first_time = true
	-- We need to create a list of possibilities from which the WML random generator would take one item with the same probability, so the items that should spawn more frequently should be more times on the list
	local function add(i)
		-- This function takes two things into consideration, the chance to drop that determines how many times it will be on the list and the problem that the first item will not be preceeded by a comma
		if var[i][2].drop_chance then
			for j = 1,var[i][2].drop_chance do
				if first_time == true then
					list = tostring(i)
					first_time = false
				else
					list = list .. "," .. tostring(i)
				end
			end
		else
			if first_time == true then
				list = tostring(i)
				first_time = false
			else
				list = list .. "," .. tostring(i)
			end
		end
	end
	for i = 1,#var do
		-- Here, we are mostly checking for level requirement and sort for the case of randomised items, and calling the previously define add() function
		if var[i][2].level <= level then
			if var[i][2].sorts_affected then
				local can_be = false
				for j in string.gmatch(var[i][2].sorts_affected, "[^%s,][^,]*") do
					if j == base_items[base_item][2].sort then
						can_be = true
					end
				end
				if can_be == true then
					add(i)
				end
			else
				add(i)
			end
		end
	end
	return helper.rand(list)
end

function wesnoth.generate_magic_item(level)
	-- This function uses the choose_item function to generate random items
	local base_item = wesnoth.choose_item(base_items,base_item,level)
	local item_code = 100 * base_item
	local has_it = helper.rand("0,0,1")	-- 33% chance that the item will have a prefix
	if has_it == 1 then
		item_code = item_code + 10000 * wesnoth.choose_item(prefixes,base_item,level) + 200000
	else
		item_code = item_code + 200000
	end
	has_it = helper.rand("0,0,1")	-- 33% chance that the item will have a suffix - therefore, the item has a 44% chance to be blank, 44% chance to have a prefix or suffix and 11% chance to have both a prefix and suffix
	if has_it == 1 then
		item_code = item_code + wesnoth.choose_item(suffixes,base_item,level)
	end
	return item_code
end

function wesnoth.generate_random_unique(level)
	-- Uniques have simply a predefined list, so choosing one is simpler than with random items
	return 60000 + wesnoth.choose_item(uniques,"item",level)
end

function wesnoth.wml_actions.drop_random_item(cfg)
	-- This is a WML tag that calls the generate_magic_item function
	local level = cfg.level or helper.wml_error("[drop_random_item] missing the required level attribute")
	local x = cfg.x or helper.wml_error("[drop_random_item] missing the required x attribute")
	local y = cfg.y or helper.wml_error("[drop_random_item] missing the required y attribute")
	local item_code = wesnoth.generate_magic_item(level)
	wesnoth.wml_actions.put_item{ x=x, y=y, id=item_code}
end

function wesnoth.wml_actions.drop_random_unique(cfg)
	-- This is a WML tag that calls the generate_random_unique function
	local level = cfg.level or helper.wml_error("[drop_random_item] missing the required level attribute")
	local x = cfg.x or helper.wml_error("[drop_random_item] missing the required x attribute")
	local y = cfg.y or helper.wml_error("[drop_random_item] missing the required y attribute")
	local item_code = wesnoth.generate_random_unique(level)
	wesnoth.wml_actions.put_item{ x=x, y=y, id=item_code}
end

function wesnoth.wml_actions.drop_random_gold(cfg)
	-- This function makes enemies drop random amounts of gold, and gives it different appearances for different amounts
	local level = cfg.level or helper.wml_error("[drop_random_gold] missing the required level attribute")
	local x = cfg.x or helper.wml_error("[drop_random_gold] missing the required x attribute")
	local y = cfg.y or helper.wml_error("[drop_random_gold] missing the required y attribute")
	local amount = helper.rand("1.." .. tostring(level*4))
	local raw_items = wesnoth.get_variable "items" or {}
	local items = {}
	items[2]=raw_items
	items[1]="items"
	items[2][#items[2]+1] = { "gold" , { amount = amount , x = x , y = y}}
	wesnoth.set_variable("items",items[2])
	local appearance
	if amount < 20 then
		appearance = "items/gold-coins-small.png"
	end
	if amount >= 20 and amout < 50 then
		appearance = "items/gold-coins-medium.png"
	end
	if amount >= 50 then
		appearance = "items/gold-coins-large.png"
	end
	wesnoth.wml_actions.item({ x=x , y=y, image=appearance, halo="halo/misc/leadership-flare-1.png:20,halo/misc/leadership-flare-2.png:20,halo/misc/leadership-flare-3.png:20,halo/misc/leadership-flare-4.png:20,halo/misc/leadership-flare-4.png:20,halo/misc/leadership-flare-6.png:20,halo/misc/leadership-flare-7.png:20,halo/misc/leadership-flare-8.png:20,halo/misc/leadership-flare-9.png:20,halo/misc/leadership-flare-10.png:20,halo/misc/leadership-flare-11.png:20,halo/misc/leadership-flare-12.png:20,halo/misc/leadership-flare-13.png:20,misc/blank-hex.png:1000" })
	wesnoth.wml_actions.event({ name = "moveto" , id="gp" .. x .. y , first_time_only= "no", { "filter", { side="1,2,3,4,5", x=x, y=y, canrecruit=true } }, { "fire_event" , { name="item_pickup", { "primary_unit" , { x=x , y=y }  } } }  })
end

function wesnoth.wml_actions.open_shop(cfg)
	-- This is similar to the inventory viewing function, just that the list of equipped items is replaced by a randomly generated list of items the trader has, and some gold costs added
	local magic_items = cfg.magical or 0
	local unique_items = cfg.unique or 0
	if magic_items == 0 and unique_items == 0 then
		helper.wml_error("No items in the shop. The number of uniques or magical items must be greater than 0.")
	end
	local price_modifier = cfg.price_modifier or 1	-- Some shops might be more expensive than others
	local level = cfg.level or helper.wml_error("[open_shop] missing the required level= attribute")
	-- Get the inventory and available gold of the unit
	local side_number = wesnoth.get_variable("side_number")
	local raw_unit = wesnoth.get_units { canrecruit=true , side=side_number }
	local unit = raw_unit[1].__cfg
	local properties = {}
	for i=1,#unit do
		if unit[i][1]=="variables" then
			properties = unit[i][2]
		end
	end
	local raw_side = wesnoth.get_sides { side=side_number }
	local side = raw_side[1].__cfg
	local original_gold = side.gold
	local result = wesnoth.synchronize_choice(
		function()
			-- Initialize variables
			local gold_trade = 0
			local bought_something = false
			local items_inventory = {}
			local items_seller = {}
			for i=1,#properties do
				if properties[i][1] == "a" then
					if (properties[i][2].id <1070000 and properties[i][2].id >= 1060000) or (properties[i][2].id <1200000 and properties[i][2].id >= 1100000) or (properties[i][2].id <2000000  and properties[i][2].id >= 1200000) then
						table.insert(items_inventory,properties[i][2].id-1000000)
					end
				end
			end
			-- Generate seller's inventory
			if unique_items then
				if unique_items > 0 then
					for i = 0,unique_items do
						table.insert(items_seller,wesnoth.generate_random_unique(level))
					end
				end
			end
			if magic_items then
				if magic_items > 0 then
					for i = 0,magic_items do
						table.insert(items_seller,wesnoth.generate_magic_item(level))
					end
				end
			end
			local dialog = { T.tooltip { id = "tooltip_large" }, T.helptip { id = "tooltip_large" },  maximum_height = 600, --maximum_width = 800,
				T.grid { T.row { T.column { T.label { id="title" , definition="title" , label="Shop"} }} ,
					T.row { T.column { T.grid {
						T.row {
							T.column { T.button { id = "gui_buy", label = "Buy" , tooltip="Buy the selected item."}},
							T.column { T.label { id = "gui_gold", label = "              " , tooltip="The gold you would have after this transaction."}},
							T.column { T.button { id = "gui_sell", label = "Sell" , tooltip="Sell the selected item."}}
						},
						T.row {
							T.column { horizontal_grow = true, T.listbox { id = "gui_seller_chosen", 
								T.list_definition { T.row { T.column { horizontal_grow = true,
									T.toggle_panel { tooltip="The list of trader's items.", T.grid { T.row {
										T.column { T.image { id = "gui_item_icon" } },
					   					T.column { T.grid {
											T.row { T.column { horizontal_alignment = "left", T.label { id = "gui_item_name" } }},
											T.row { T.column { horizontal_alignment = "left", T.label { id = "gui_item_price" }} }
										}}
									}}}
								}}}
							}},
							T.column {T.label { id = "gui_description", label = " " , tooltip="Description of the selected item."}},
							T.column { horizontal_grow = true, T.listbox { id = "gui_inventory_chosen", 
								T.list_definition { T.row { T.column { horizontal_grow = true,
									T.toggle_panel { tooltip="The list of items in inventory.", T.grid { T.row {
										T.column { T.image { id = "gui_item_icon" } },
					   					T.column { T.grid {
											T.row { T.column { horizontal_alignment = "left", T.label { id = "gui_item_name" }} },
											T.row { T.column { horizontal_alignment = "left", T.label { id = "gui_item_price" } }}
										}}
									}}}
								}}}
							}}
						}
					}}},
					T.row { T.column { T.grid { T.row {
						T.column { T.button { id = "ok", label = "Agree with trade" , tooltip="Do all the changes you selected. If you don't bought anything, you'll pay an amount of gold equal to character level as a refund for annoying the trader." } },
						T.column { T.button { id = "cancel", label = "Disagree with trade", tooltip="Undo all selected changes, paying an amount of gold equal to character level as a refund for annoying the trader." } }
					}}}}
				}}
			-- Same as in inventory manipulation
			local function preshow()
				local function select_seller()
					for i=1,(#items_inventory + #items_seller) do
						wesnoth.set_dialog_value("                               ", "gui_seller_chosen", i, "gui_item_name")
						wesnoth.set_dialog_value("                               ", "gui_seller_chosen", i, "gui_item_price")
						wesnoth.set_dialog_value( "misc/blank-hex.png", "gui_seller_chosen", i, "gui_item_icon")
					end
					for i=1,(#items_inventory + #items_seller) do
						wesnoth.set_dialog_value("                               ", "gui_inventory_chosen", i, "gui_item_name")
						wesnoth.set_dialog_value("                               ", "gui_inventory_chosen", i, "gui_item_price")
						wesnoth.set_dialog_value("misc/blank-hex.png", "gui_inventory_chosen", i, "gui_item_icon")
					end
					for i=1,#items_seller do
						local description,item_name,item_type,item_appearance,level_req,str_req,dex_req,int_req,sell_price,buy_price = wesnoth.describe_item(items_seller[i])
						wesnoth.set_dialog_value( item_name , "gui_seller_chosen", i, "gui_item_name")
						wesnoth.set_dialog_value( "Price: " .. tostring((buy_price * price_modifier) - (buy_price * price_modifier) % 1) , "gui_seller_chosen", i, "gui_item_price")
						wesnoth.set_dialog_value( item_appearance , "gui_seller_chosen", i, "gui_item_icon")
					end
					for i=1,#items_inventory do
						local description,item_name,item_type,item_appearance,level_req,str_req,dex_req,int_req,sell_price = wesnoth.describe_item(items_inventory[i])
						wesnoth.set_dialog_value( item_name , "gui_inventory_chosen", i, "gui_item_name")
						wesnoth.set_dialog_value( "Sell value: " .. tostring(sell_price) , "gui_inventory_chosen", i, "gui_item_price")
						wesnoth.set_dialog_value( item_appearance , "gui_inventory_chosen", i, "gui_item_icon")
					end
					local preshow_seller_selected = wesnoth.get_dialog_value "gui_seller_chosen"
					if items_seller[preshow_seller_selected] then
						wesnoth.set_dialog_value("--------------------------------------\n" .. wesnoth.describe_item(items_seller[preshow_seller_selected]) , "gui_description")
					end
					wesnoth.set_dialog_value( "Gold: " .. original_gold + gold_trade , "gui_gold")
				end
				local function select_inventory()
					for i=1,(#items_inventory + #items_seller) do
						wesnoth.set_dialog_value("                               ", "gui_seller_chosen", i, "gui_item_name")
						wesnoth.set_dialog_value("                               ", "gui_seller_chosen", i, "gui_item_price")
						wesnoth.set_dialog_value( "misc/blank-hex.png", "gui_seller_chosen", i, "gui_item_icon")
					end
					for i=1,(#items_inventory + #items_seller) do
						wesnoth.set_dialog_value("                               ", "gui_inventory_chosen", i, "gui_item_name")
						wesnoth.set_dialog_value("                               ", "gui_inventory_chosen", i, "gui_item_price")
						wesnoth.set_dialog_value("misc/blank-hex.png", "gui_inventory_chosen", i, "gui_item_icon")
					end
					for i=1,#items_seller do
						local description,item_name,item_type,item_appearance,level_req,str_req,dex_req,int_req,sell_price,buy_price = wesnoth.describe_item(items_seller[i])
						wesnoth.set_dialog_value( item_name , "gui_seller_chosen", i, "gui_item_name")
						wesnoth.set_dialog_value( "Price: " .. tostring((buy_price * price_modifier) - (buy_price * price_modifier) % 1) , "gui_seller_chosen", i, "gui_item_price")
						wesnoth.set_dialog_value( item_appearance , "gui_seller_chosen", i, "gui_item_icon")
					end
					for i=1,#items_inventory do
						local description,item_name,item_type,item_appearance,level_req,str_req,dex_req,int_req,sell_price = wesnoth.describe_item(items_inventory[i])
						wesnoth.set_dialog_value( item_name , "gui_inventory_chosen", i, "gui_item_name")
						wesnoth.set_dialog_value( "Sell value: " .. tostring(sell_price) , "gui_inventory_chosen", i, "gui_item_price")
						wesnoth.set_dialog_value( item_appearance , "gui_inventory_chosen", i, "gui_item_icon")
					end
					local preshow_inventory_selected = wesnoth.get_dialog_value "gui_inventory_chosen"
					if items_inventory[preshow_inventory_selected] then
						wesnoth.set_dialog_value("--------------------------------------\n" .. wesnoth.describe_item(items_inventory[preshow_inventory_selected]) , "gui_description")
					end
					wesnoth.set_dialog_value( "Gold: " .. original_gold + gold_trade , "gui_gold")
				end
				local function sell()
					local preshow_inventory_selected = wesnoth.get_dialog_value "gui_inventory_chosen"
					if items_inventory[preshow_inventory_selected] then
						local description2,item_name2,item_type2,item_appearance,level_req,str_req,dex_req,int_req,sell_price = wesnoth.describe_item(items_inventory[preshow_inventory_selected])
						gold_trade = gold_trade + sell_price
						items_seller[#items_seller+1] = items_inventory[preshow_inventory_selected]
						table.remove(items_inventory,preshow_inventory_selected)
						select_seller()
					end
				end
				local function buy()
					-- Like in inventory manipulation, but we are checking if the player has enough gold
					local preshow_seller_selected = wesnoth.get_dialog_value "gui_seller_chosen"
					if items_seller[preshow_seller_selected] then
						local description2,item_name2,item_type2,item_appearance,level_req,str_req,dex_req,int_req,sell_price,buy_price = wesnoth.describe_item(items_seller[preshow_seller_selected])
						if original_gold + gold_trade >= (buy_price * price_modifier) - (buy_price * price_modifier) % 1 then
							gold_trade = gold_trade - (buy_price * price_modifier) - (buy_price * price_modifier) % 1
							items_inventory[#items_inventory+1] = items_seller[preshow_seller_selected]
							table.remove(items_seller,preshow_seller_selected)
							bought_something = true
							select_inventory()
						else
							wesnoth.message_box("You can't afford this item.")
						end
					end
				end
				wesnoth.set_dialog_callback(select_seller, "gui_seller_chosen")
				wesnoth.set_dialog_callback(select_inventory, "gui_inventory_chosen")
				wesnoth.set_dialog_callback(buy, "gui_buy")
				wesnoth.set_dialog_callback(sell, "gui_sell")
				select_seller()
				wesnoth.set_dialog_value("--------------------------------------\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n" , "gui_description")
			end

			local function postshow()
			end

			local choice_done = wesnoth.show_dialog(dialog, preshow, postshow)
			return {choice_done = choice_done, inventory = table.concat(items_inventory,",") , seller = table.concat(items_seller,",") , gold_trade = gold_trade}

		end
	)
	local items_inventory = {}
	for j in string.gmatch( result.inventory , "[^%s,][^,]*") do
		items_inventory[#items_inventory+1] = tonumber(j)
	end
	-- Items seller is returned only because it might be needed later
	if result.choice_done == -1 then
		-- The manipulation of the [a] tag code is easier here, because we are manipulating only the inventory, therefore I found it better to simply delete all inventory notes and base them
		-- on the ones that the trade menu returned
		local i = 1
		while i <= #properties do
			if properties[i] then
				if properties[i][1] == "a" then
					if (properties[i][2].id <1070000 and properties[i][2].id >= 1060000) or (properties[i][2].id <1200000 and properties[i][2].id >= 1100000) or (properties[i][2].id <2000000 and properties[i][2].id >= 1200000) then
						table.remove(properties,i)
						i = i - 1
					end
				end
			end
			i = i + 1
		end
		for i = 1,#items_inventory do
			table.insert(properties,{ "a" , { id = 1000000 + items_inventory[i]}})
		end
		for i=1,#unit do
			if unit[i][1]=="variables" then
				unit[i][2] = properties
			end
		end
		unit = wesnoth.update_stats(unit)
		wesnoth.put_unit(unit)
		-- Paying gold, negative gold_trade means that the player lost money in the trade, probably bought expensive stuff
		wesnoth.wml_actions.gold{ side = side_number , amount = result.gold_trade}
	end
	if bought_something == false then
		wesnoth.wml_actions.gold{ side = side_number , amount = -1 * unit.level}	-- This is just to make sure that the player won't keep opening the shop until he finds what he wants
	end

end
