local helper = wesnoth.require "lua/helper.lua"

function wesnoth.wml_actions.spawn_on_approach(cfg)
	local types = cfg.types or helper.wml_error("[spawn_on_approach] missing the required types= attribute")
	local quantity = cfg.quantity or 1
	local x = cfg.x or helper.wml_error("[spawn_on_approach] missing the required x= attribute")
	local y = cfg.y or helper.wml_error("[spawn_on_approach] missing the required y= attribute")
	local side = cfg.side or helper.wml_error("[spawn_on_approach] missing the required side= attribute")
	local event = { name = "moveto"}
	if cfg.id then
		event.id = cfg.id .. tostring(x) .. "," .. tostring(y)
	end
	local x1 = x - 12
	local x2 = x + 12
	local y1 = y - 12
	local y2 = y + 12
	if x1 < 1 then x1 = 1 end
	if y1 < 1 then y1 = 1 end
	event[1] = {"filter" , { side="1,2,3,4,5" , x = tostring(x1) .. "-" .. tostring (x2) , y = tostring(y1) .. "-" .. tostring(y2)}}
	if cfg.boss then
		event[2] = { "unit" , { side = side, type = boss , x = x , y = y , random_gender = true, random_traits = true, generate_name = true }}
	end
	event[#event+1] = { "set_variable" , { name = "count" , value = quantity }}
	event[#event+1] = { "while" , {
			{"variable", { name = "count" , greater_than = 0 }},
			{"do" , {
				{ "set_variable" , { name = "type" , rand = types }},
				{ "unit" , { side = side, type = "$type" , x = x , y = y , random_gender = true, random_traits = true, generate_name = true }},
				{ "set_variable" , { name = "count" , sub = 1 }}
			}}
		}}
	event[#event+1] = { "clear_variable" , { name = "count,type" }}
	wesnoth.wml_actions.event(event)

end

function wesnoth.wml_actions.spawn_constantly(cfg)
	local types = cfg.types or helper.wml_error("[spawn_constantly] missing the required types= attribute")
	local period = cfg.period or 1
	local x = cfg.x or helper.wml_error("[spawn_constantly] missing the required x= attribute")
	local y = cfg.y or helper.wml_error("[spawn_constantly] missing the required y= attribute")
	local side = cfg.side or helper.wml_error("[spawn_constantly] missing the required side= attribute")
	local event = { name = "new_turn" , first_time_only = false}
	if cfg.id then
		event.id = cfg.id .. tostring(x) .. "," .. tostring(y)
	end
	local x1 = x - 12
	local x2 = x + 12
	local y1 = y - 12
	local y2 = y + 12
	if x1 < 1 then x1 = 1 end
	if y1 < 1 then y1 = 1 end
	event[1] = {"filter_condition" , { { "not" ,{ { "have_unit" ,{
			side="1,2,3,4,5" , x = tostring(x1) .. "-" .. tostring (x2) , y = tostring(y1) .. "-" .. tostring(y2)
		}} }} }}
	event[2] = { "set_variable" , { name = "do_spawn" , to_variable = "turn_number" }}
	event[3] = { "set_variable" , { name = "do_spawn" , modulo = period }}
	event[4] = { "if" , {
			{"variable", { name = "do_spawn" , equals = 0 }},
			{"then" , {
				{ "set_variable" , { name = "type" , rand = types }},
				{ "unit" , { side = side, type = "$type" , x = x , y = y , random_gender = true, random_traits = true, generate_name = true }}
			}}
		}}
	event[5] = { "clear_variable" , { name = "do_spawn,type" }}
	wesnoth.wml_actions.event(event)

end

function wesnoth.wml_actions.delete_spawn_point(cfg)
	local id = cfg.id or helper.wml_error("[delete_spawn_point] missing the required id= attribute")
	local x = cfg.x or helper.wml_error("[delete_spawn_point] missing the required x= attribute")
	local y = cfg.y or helper.wml_error("[delete_spawn_point] missing the required y= attribute")
	wesnoth.wml_actions.event({ id = id .. tostring(x) .. "," .. tostring(y) ,  remove = true})
end
